from typing import Literal

import rapidfuzz


def compare_strings(
    string1: str,
    string2: str,
    similarity_measure: Literal[
        "JaroWinkler", "Levenshtein", "DamerauLevenshtein"
    ] = "JaroWinkler",
) -> float:
    """
    Returns the normalized similarity measure between string1 and string2, as
    determined by the similarlity measure. The higher the normalized similarity measure
    (up to 1.0), the more similar string1 and string2 are. A normalized similarity
    measure of 0.0 means string1 and string 2 are not at all similar. This function
    expects basic text cleaning (e.g. removal of numeric characters, trimming of spaces,
    etc.) to already have been performed on the input strings.

    :param string1: First string for comparison.
    :param string2: Second string for comparison.
    :param similarity_measure: The method used to measure the similarity between two
        strings, defaults to "JaroWinkler".
     - JaroWinkler: a ratio of matching characters and transpositions needed to
        transform string1 into string2.
     - Levenshtein: the number of edits (excluding transpositions) needed to transform
        string1 into string2.
     - DamerauLevenshtein: the number of edits (including transpositions) needed to
        transform string1 into string2.
    :return: The normalized similarity between string1 and string2, with 0 representing
        no similarity between string1 and string2, and 1 meaning string1 and string2 are
        dentical words.
    """
    if similarity_measure == "JaroWinkler":
        return rapidfuzz.distance.JaroWinkler.normalized_similarity(string1, string2)
    elif similarity_measure == "Levenshtein":
        return rapidfuzz.distance.Levenshtein.normalized_similarity(string1, string2)
    elif similarity_measure == "DamerauLevenshtein":
        return rapidfuzz.distance.DamerauLevenshtein.normalized_similarity(
            string1, string2
        )



import io
from typing import Iterator
from typing import TextIO
from typing import Tuple

from azure.storage.blob import download_blob_from_url

from phdi.cloud.azure import AzureCredentialManager


def download_from_fhir_export_response(
    export_response: dict,
    cred_manager: AzureCredentialManager,
) -> Iterator[Tuple[str, TextIO]]:
    """
    Accepts the export response content as specified here:
    https://hl7.org/fhir/uv/bulkdata/export/index.html#response---complete-status

    Loops through the "output" array and yields the resource_type (e.g., Patient)
    along with TextIO wrapping ndjson content.

    :param export_response: A dictionary holding the final export response.
    :param cred_manager: The credential manager used to authenticate to the
      storage account.
    :return: An iterator of tuples. Each tuple is comprised of:

      * FHIR resource type (str)
      * Export file content (io.TextIO)
    """
    # TODO: Handle error array that could be contained in the response content.

    for export_entry in export_response.get("output", []):
        resource_type = export_entry.get("type")
        blob_url = export_entry.get("url")

        yield (
            resource_type,
            _download_export_blob(blob_url=blob_url, cred_manager=cred_manager),
        )


def _download_export_blob(
    blob_url: str, cred_manager: AzureCredentialManager, encoding: str = "utf-8"
) -> TextIO:
    """
    Downloads an export file blob.

    :param blob_url: The blob URL location to download from blob storage.
    :param cred_manager: The credential manager used to authenticate to the
      storage account.
    :param encoding: The character encoding to apply to the downloaded content.
      Default: utf-8
    :return: Content of export file located at `blob_url`
    """
    bytes_buffer = io.BytesIO()
    azure_creds = cred_manager.get_credential_object()
    download_blob_from_url(
        blob_url=blob_url, output=bytes_buffer, credential=azure_creds
    )
    text_buffer = io.TextIOWrapper(buffer=bytes_buffer, encoding=encoding, newline="\n")
    text_buffer.seek(0)

    return text_buffer



import unicodedata

"""
The Double Metaphone phonetic encoding algorithm is an improvement to the
phonetic representation algorithm Soundex, commonly used to identify what an
English name "sounds like." However, rather than produce an exact phonetic
description of a name, Double Metaphone produces an intentionally approximate
phonetic representation, in order to account for the way speakers vary their
pronunciations and misspell or otherwise vary words and names. It also
returns both a primary and a secondary encoding for a string (hence the
name "Double" Metaphone), which accounts for some ambiguous cases as well as
for multiple variants of surnames with common ancestry. Double Metaphone is
capable of accounting for irregularities in the spelling and phonetics of
English, Slavic, Germanic, Celtic, Greek, French, Italian, Spanish, Chinese,
and more.
"""

VOWELS = ["A", "E", "I", "O", "U", "Y"]
SILENT_STARTERS = ["GN", "KN", "PN", "WR", "PS"]


class DoubleMetaphone(object):
    """
    A Double Metaphone parsing object capable of performing the double
    coding algorithm on a given input string.
    """

    def __init__(self):
        self.position = 0
        self.primary_phone = ""
        self.secondary_phone = ""
        # next is set to a tuple of the next characters in the primary and
        # secondary codes and to indicate how many characters to move forward
        # in the string.  The secondary code letter is given only when it is
        # different than the primary.
        self.next = (None, 1)

    def __call__(self, string: str):
        return self.parse(string)

    def check_word_start(self):
        # Skip silent letters when they start a word (because they're not
        # phonetic if they're silent)
        if self.word.get_letters(0, 2) in SILENT_STARTERS:
            self.position += 1
        # Starting with 'X' frequently pronounced 'Z' e.g. 'Xavier'
        if self.word.get_letters(0) == "X":
            # 'Z' maps to 'S'
            self.primary_phone = self.secondary_phone = "S"
            self.position += 1

    def process_vowel(self):
        self.next = (None, 1)

        # All starting vowels map to 'A'
        if self.position == self.word.start_index:
            self.next = ("A", 1)

    def process_b(self):
        # "-mb", e.g., "dumb", already accounted for, see 'M'
        if self.word.buffer[self.position + 1] == "B":
            self.next = ("P", 2)
        else:
            self.next = ("P", 1)

    def process_c(self):
        buffer = self.word.buffer
        position = self.position
        start_index = self.word.start_index

        # Handle some germanic linguistics
        if (
            position > start_index + 1
            and buffer[position - 2] not in VOWELS
            and buffer[position - 1 : self.position + 2] == "ACH"
            and buffer[position + 2] not in ["I"]
            and (
                buffer[position + 2] not in ["E"]
                or buffer[position - 2 : position + 4] in ["BACHER", "MACHER"]
            )
        ):
            self.next = ("K", 2)

        # Special case 'CAESAR'
        elif (
            position == start_index
            and buffer[start_index : start_index + 6] == "CAESAR"
        ):
            self.next = ("S", 2)

        # Italian 'chianti'
        elif buffer[position : position + 4] == "CHIA":
            self.next = ("K", 2)
        elif buffer[position : position + 2] == "CH":
            # e.g. 'michael'
            if position > start_index and buffer[position : position + 4] == "CHAE":
                self.next = ("K", "X", 2)
            elif (
                position == start_index
                and (
                    buffer[position + 1 : position + 6] in ["HARAC", "HARIS"]
                    or buffer[position + 1 : position + 4]
                    in ["HOR", "HYM", "HIA", "HEM"]
                )
                and buffer[start_index : start_index + 5] != "CHORE"
            ):
                self.next = ("K", 2)

            # Germanic, greek, or other phonetic 'ch' for 'kh' sound
            elif (
                buffer[start_index : start_index + 4] in ["VAN ", "VON "]
                or buffer[start_index : start_index + 3] == "SCH"
                or buffer[position - 2 : position + 4] in ["ORCHES", "ARCHIT", "ORCHID"]
                or buffer[position + 2] in ["T", "S"]
                or (
                    (
                        buffer[position - 1] in ["A", "O", "U", "E"]
                        or position == start_index
                    )
                    and (
                        buffer[position + 2]
                        in ["L", "R", "N", "M", "B", "H", "F", "V", "W", " "]
                    )
                )
            ):
                self.next = ("K", 2)
            else:
                if position > start_index:
                    if buffer[start_index : start_index + 2] == "MC":
                        self.next = ("K", 2)
                    else:
                        self.next = ("X", "K", 2)
                else:
                    self.next = ("X", 2)

        # e.g, 'czerny'
        elif (
            buffer[position : position + 2] == "CZ"
            and buffer[position - 2 : position + 2] != "WICZ"
        ):
            self.next = ("S", "X", 2)

        # e.g., 'focaccia'
        elif buffer[position + 1 : position + 4] == "CIA":
            self.next = ("X", 3)

        # Double 'C', but not if e.g. 'McClellan'
        elif buffer[position : position + 2] == "CC" and not (
            position == (start_index + 1) and buffer[start_index] == "M"
        ):
            # 'bellocchio' but not 'bacchus'
            if (
                buffer[position + 2] in ["I", "E", "H"]
                and buffer[position + 2 : position + 4] != "HU"
            ):
                # 'accident', 'accede' 'succeed'
                if (
                    position == (start_index + 1) and buffer[start_index] == "A"
                ) or buffer[position - 1 : position + 4] in ["UCCEE", "UCCES"]:
                    self.next = ("KS", 3)

                # 'bacci', 'bertucci', other italian
                else:
                    self.next = ("X", 3)
            else:
                self.next = ("K", 2)

        elif buffer[position : position + 2] in ["CK", "CG", "CQ"]:
            self.next = ("K", 2)
        elif buffer[position : position + 2] in ["CI", "CE", "CY"]:
            # Italian vs. English
            if buffer[position : position + 3] in ["CIO", "CIE", "CIA"]:
                self.next = ("S", "X", 2)
            else:
                self.next = ("S", 2)
        else:
            # Name set in 'mac caffrey', 'mac gregor'
            if buffer[position + 1 : position + 3] in [" C", " Q", " G"]:
                self.next = ("K", 3)
            else:
                if buffer[position + 1] in ["C", "K", "Q"] and buffer[
                    position + 1 : position + 3
                ] not in ["CE", "CI"]:
                    self.next = ("K", 2)

                # Default sound for 'C'
                else:
                    self.next = ("K", 1)

    def process_d(self):
        if self.word.buffer[self.position : self.position + 2] == "DG":
            # e.g. 'edge'
            if self.word.buffer[self.position + 2] in ["I", "E", "Y"]:
                self.next = ("J", 3)
            else:
                self.next = ("TK", 2)
        elif self.word.buffer[self.position : self.position + 2] in ["DT", "DD"]:
            self.next = ("T", 2)
        else:
            self.next = ("T", 1)

    def process_f(self):
        if self.word.buffer[self.position + 1] == "F":
            self.next = ("F", 2)
        else:
            self.next = ("F", 1)

    def process_g(self):
        buffer = self.word.buffer
        position = self.position
        start_index = self.word.start_index
        if buffer[position + 1] == "H":
            if position > start_index and buffer[position - 1] not in VOWELS:
                self.next = ("K", 2)
            elif position < (start_index + 3):
                # 'Ghislane', 'Ghiradelli'
                if position == start_index:
                    if buffer[position + 2] == "I":
                        self.next = ("J", 2)
                    else:
                        self.next = ("K", 2)

            # Parker's rule for e.g., 'Hugh'
            elif (
                (
                    position > (start_index + 1)
                    and buffer[position - 2] in ["B", "H", "D"]
                )
                or (
                    position > (start_index + 2)
                    and buffer[position - 3] in ["B", "H", "D"]
                )
                or (position > (start_index + 3) and buffer[position - 4] in ["B", "H"])
            ):
                self.next = (None, 2)

            else:
                # e.g., 'laugh', 'McLaughlin', 'cough', 'gough', 'rough', 'tough'
                if (
                    position > (start_index + 2)
                    and buffer[position - 1] == "U"
                    and buffer[position - 3] in ["C", "G", "L", "R", "T"]
                ):
                    self.next = ("F", 2)
                else:
                    if position > start_index and buffer[position - 1] != "I":
                        self.next = ("K", 2)

        elif buffer[position + 1] == "N":
            if (
                position == (start_index + 1)
                and buffer[start_index] in VOWELS
                and not self.word.is_slavo_germanic
            ):
                self.next = ("KN", "N", 2)

            else:
                # Not e.g. 'cagney'
                if (
                    buffer[position + 2 : position + 4] != "EY"
                    and buffer[position + 1] != "Y"
                    and not self.word.is_slavo_germanic
                ):
                    self.next = ("N", "KN", 2)
                else:
                    self.next = ("KN", 2)

        # e.g. 'Tagliaro'
        elif (
            buffer[position + 1 : position + 3] == "LI"
            and not self.word.is_slavo_germanic
        ):
            self.next = ("KL", "L", 2)

        # -ges-,-gep-,-gel-, -gie- at beginning
        elif position == start_index and (
            buffer[position + 1] == "Y"
            or buffer[position + 1 : position + 3]
            in ["ES", "EP", "EB", "EL", "EY", "IB", "IL", "IN", "IE", "EI", "ER"]
        ):
            self.next = ("K", "J", 2)

        # -ger-,  -gy-
        elif (
            (buffer[position + 1 : position + 3] == "ER" or buffer[position + 1] == "Y")
            and buffer[start_index : start_index + 6]
            not in ["DANGER", "RANGER", "MANGER"]
            and buffer[position - 1] not in ["E", "I"]
            and buffer[position - 1 : position + 2] not in ["RGY", "OGY"]
        ):
            self.next = ("K", "J", 2)

        # Italian e.g, 'Biaggi'
        elif buffer[position + 1] in ["E", "I", "Y"] or buffer[
            position - 1 : position + 3
        ] in ["AGGI", "OGGI"]:
            # Germanic
            if (
                buffer[start_index : start_index + 4] in ["VON ", "VAN "]
                or buffer[start_index : start_index + 3] == "SCH"
                or buffer[position + 1 : position + 3] == "ET"
            ):
                self.next = ("K", 2)

            else:
                # Always soft if following the French ending
                if buffer[position + 1 : position + 5] == "IER ":
                    self.next = ("J", 2)
                else:
                    self.next = ("J", "K", 2)

        elif buffer[position + 1] == "G":
            self.next = ("K", 2)
        else:
            self.next = ("K", 1)

    def process_h(self):
        # Keep if self.word.start_index & before vowel, or between 2 vowels
        if (
            self.position == self.word.start_index
            or self.word.buffer[self.position - 1] in VOWELS
        ) and self.word.buffer[self.position + 1] in VOWELS:
            self.next = ("H", 2)
        else:
            self.next = (None, 1)

    def process_j(self):
        buffer = self.word.buffer
        position = self.position
        start_index = self.word.start_index

        # Spanish, 'Jose', 'Jacinto', etc.
        if (
            buffer[self.position : self.position + 4] == "JOSE"
            or buffer[start_index : start_index + 4] == "SAN "
        ):
            if (position == start_index and buffer[position + 4] == " ") or buffer[
                start_index : start_index + 4
            ] == "SAN ":
                self.next = ("H",)
            else:
                self.next = ("J", "H")

        # Yankelovich/Jankelowicz
        elif (
            position == start_index
            and buffer[self.position : self.position + 4] != "JOSE"
        ):
            self.next = ("J", "A")

        else:
            # Spanish pronunciation of e.g. 'Bajador'
            if (
                buffer[position - 1] in VOWELS
                and not self.word.is_slavo_germanic
                and buffer[position + 1] in ["A", "O"]
            ):
                self.next = ("J", "H")
            else:
                if position == self.word.end_index:
                    self.next = ("J", " ")
                else:
                    if buffer[position + 1] not in [
                        "L",
                        "T",
                        "K",
                        "S",
                        "N",
                        "M",
                        "B",
                        "Z",
                    ] and buffer[position - 1] not in ["S", "K", "L"]:
                        self.next = ("J",)
                    else:
                        self.next = (None,)

        if buffer[position + 1] == "J":
            self.next = self.next + (2,)
        else:
            self.next = self.next + (1,)

    def process_k(self):
        if self.word.buffer[self.position + 1] == "K":
            self.next = ("K", 2)
        else:
            self.next = ("K", 1)

    def process_l(self):
        buffer = self.word.buffer
        position = self.position
        end_index = self.word.end_index

        if buffer[position + 1] == "L":
            # Spanish e.g. 'cabrillo', 'gallegos'
            if (
                position == (end_index - 2)
                and buffer[position - 1 : position + 3] in ["ILLO", "ILLA", "ALLE"]
            ) or (
                (
                    buffer[end_index - 1 : end_index + 1] in ["AS", "OS"]
                    or buffer[end_index] in ["A", "O"]
                )
                and buffer[position - 1 : position + 3] == "ALLE"
            ):
                self.next = ("L", "", 2)
            else:
                self.next = ("L", 2)
        else:
            self.next = ("L", 1)

    def process_m(self):
        buffer = self.word.buffer
        position = self.position
        if (
            buffer[position + 1 : position + 4] == "UMB"
            and (
                position + 1 == self.word.end_index
                or buffer[position + 2 : position + 4] == "ER"
            )
        ) or buffer[position + 1] == "M":
            self.next = ("M", 2)
        else:
            self.next = ("M", 1)

    def process_n(self):
        if self.word.buffer[self.position + 1] == "N":
            self.next = ("N", 2)
        else:
            self.next = ("N", 1)

    def process_p(self):
        if self.word.buffer[self.position + 1] == "H":
            self.next = ("F", 2)

        # Account for "Campbell", "raspberry"
        elif self.word.buffer[self.position + 1] in ["P", "B"]:
            self.next = ("P", 2)
        else:
            self.next = ("P", 1)

    def process_q(self):
        if self.word.buffer[self.position + 1] == "Q":
            self.next = ("K", 2)
        else:
            self.next = ("K", 1)

    def process_r(self):
        buffer = self.word.buffer
        position = self.position
        end_index = self.word.end_index

        # French e.g. 'Rogier', but exclude e.g. 'Hochmeier'
        if (
            position == end_index
            and not self.word.is_slavo_germanic
            and buffer[position - 2 : position] == "IE"
            and buffer[position - 4 : position - 2] not in ["ME", "MA"]
        ):
            self.next = ("", "R")
        else:
            self.next = ("R",)
        if buffer[position + 1] == "R":
            self.next = self.next + (2,)
        else:
            self.next = self.next + (1,)

    def process_s(self):
        buffer = self.word.buffer
        position = self.position
        start_index = self.word.start_index
        end_index = self.word.end_index

        # Special cases like 'Carlisle', 'Carlysle'
        if buffer[position - 1 : position + 2] in ["ISL", "YSL"]:
            self.next = (None, 1)

        # special case 'sugar-'
        elif (
            position == start_index and buffer[start_index : start_index + 5] == "SUGAR"
        ):
            self.next = ("X", "S", 1)

        elif buffer[position : position + 2] == "SH":
            # Germanic
            if buffer[position + 1 : position + 5] in ["HEIM", "HOEK", "HOLM", "HOLZ"]:
                self.next = ("S", 2)
            else:
                self.next = ("X", 2)

        # Italian & Armenian
        elif (
            buffer[position : position + 3] in ["SIO", "SIA"]
            or buffer[position : position + 4] == "SIAN"
        ):
            if not self.word.is_slavo_germanic:
                self.next = ("S", "X", 3)
            else:
                self.next = ("S", 3)

        # Lots of anglicisations to match here
        # Germanic, e.g. 'Smith' == 'Schmidt', 'Snider' == 'Schneider'
        # Slavic languages, catch -sz- and the tricky Hungarian
        # pronunciation with an 's'
        elif (
            position == start_index and buffer[position + 1] in ["M", "N", "L", "W"]
        ) or buffer[position + 1] == "Z":
            self.next = ("S", "X")
            if buffer[position + 1] == "Z":
                self.next = self.next + (2,)
            else:
                self.next = self.next + (1,)

        elif buffer[position : position + 2] == "SC":
            # Schlesinger's rule
            if buffer[position + 2] == "H":
                # Dutch origin, e.g. 'school', 'schooner'
                if buffer[position + 3 : position + 5] in [
                    "OO",
                    "ER",
                    "EN",
                    "UY",
                    "ED",
                    "EM",
                ]:
                    # 'Schermerhorn', 'Schenker'
                    if buffer[position + 3 : position + 5] in ["ER", "EN"]:
                        self.next = ("X", "SK", 3)
                    else:
                        self.next = ("SK", 3)

                else:
                    if (
                        position == start_index
                        and buffer[start_index + 3] not in VOWELS
                        and buffer[start_index + 3] != "W"
                    ):
                        self.next = ("X", "S", 3)
                    else:
                        self.next = ("X", 3)

            elif buffer[position + 2] in ["I", "E", "Y"]:
                self.next = ("S", 3)
            else:
                self.next = ("SK", 3)

        # French e.g. 'Resnais', 'Artois'
        elif position == end_index and buffer[position - 2 : position] in ["AI", "OI"]:
            self.next = ("", "S", 1)
        else:
            self.next = ("S",)
            if buffer[position + 1] in ["S", "Z"]:
                self.next = self.next + (2,)
            else:
                self.next = self.next + (1,)

    def process_t(self):
        buffer = self.word.buffer
        position = self.position
        start_index = self.word.start_index

        if buffer[position : position + 4] == "TION":
            self.next = ("X", 3)
        elif buffer[position : position + 3] in ["TIA", "TCH"]:
            self.next = ("X", 3)
        elif (
            buffer[position : position + 2] == "TH"
            or buffer[position : position + 3] == "TTH"
        ):
            # 'Thomas', 'Thames' or hard-sound Germanic
            if (
                buffer[position + 2 : position + 4] in ["OM", "AM"]
                or buffer[start_index : start_index + 4] in ["VON ", "VAN "]
                or buffer[start_index : start_index + 3] == "SCH"
            ):
                self.next = ("T", 2)
            else:
                self.next = ("0", "T", 2)

        elif buffer[position + 1] in ["T", "D"]:
            self.next = ("T", 2)
        else:
            self.next = ("T", 1)

    def process_v(self):
        if self.word.buffer[self.position + 1] == "V":
            self.next = ("F", 2)
        else:
            self.next = ("F", 1)

    def process_w(self):
        buffer = self.word.buffer
        position = self.position
        start_index = self.word.start_index

        if buffer[position : position + 2] == "WR":
            self.next = ("R", 2)
        elif position == start_index and (
            buffer[position + 1] in VOWELS or buffer[position : position + 2] == "WH"
        ):
            # Wasserman should match Vasserman
            if buffer[position + 1] in VOWELS:
                self.next = ("A", "F", 1)
            else:
                self.next = ("A", 1)

        # Arnow should match Arnoff
        elif (
            (position == self.word.end_index and buffer[position - 1] in VOWELS)
            or buffer[position - 1 : position + 4]
            in ["EWSKI", "EWSKY", "OWSKI", "OWSKY"]
            or buffer[start_index : start_index + 3] == "SCH"
        ):
            self.next = ("", "F", 1)

        # Polish e.g. 'Filipowicz'
        elif buffer[position : position + 4] in ["WICZ", "WITZ"]:
            self.next = ("TS", "FX", 4)

        # By default, 'W' is skipped phonetically
        else:
            self.next = (None, 1)

    def process_x(self):
        buffer = self.word.buffer
        position = self.position

        # French e.g. 'Breaux'
        self.next = (None,)
        if not (
            position == self.word.end_index
            and (
                buffer[position - 3 : position] in ["IAU", "EAU"]
                or buffer[position - 2 : position] in ["AU", "OU"]
            )
        ):
            self.next = ("KS",)

        if buffer[position + 1] in ["C", "X"]:
            self.next = self.next + (2,)
        else:
            self.next = self.next + (1,)

    def process_z(self):
        # Chinese pinyin e.g. 'zhao'
        if self.word.buffer[self.position + 1] == "H":
            self.next = ("J",)
        elif self.word.buffer[self.position + 1 : self.position + 3] in [
            "ZO",
            "ZI",
            "ZA",
        ] or (
            self.word.is_slavo_germanic
            and self.position > self.word.start_index
            and self.word.buffer[self.position - 1] != "T"
        ):
            self.next = ("S", "TS")
        else:
            self.next = ("S",)
        if (
            self.word.buffer[self.position + 1] == "Z"
            or self.word.buffer[self.position + 1] == "H"
        ):
            self.next = self.next + (2,)
        else:
            self.next = self.next + (1,)

    def parse(self, input):
        self.word = Word(input)
        self.position = self.word.start_index
        self.primary_phone = ""
        self.secondary_phone = ""
        self.next = (None, 1)
        self.check_word_start()

        # Loop through chars in word.buffer
        while self.position <= self.word.end_index:
            character = self.word.buffer[self.position]

            # Vowels and spaces get special handling
            if character in VOWELS:
                self.process_vowel()
            elif character == " ":
                self.position += 1
                continue

            # Other letters each have their own function
            elif character == "B":
                self.process_b()
            elif character == "C":
                self.process_c()
            elif character == "D":
                self.process_d()
            elif character == "F":
                self.process_f()
            elif character == "G":
                self.process_g()
            elif character == "H":
                self.process_h()
            elif character == "J":
                self.process_j()
            elif character == "K":
                self.process_k()
            elif character == "L":
                self.process_l()
            elif character == "M":
                self.process_m()
            elif character == "N":
                self.process_n()
            elif character == "P":
                self.process_p()
            elif character == "Q":
                self.process_q()
            elif character == "R":
                self.process_r()
            elif character == "S":
                self.process_s()
            elif character == "T":
                self.process_t()
            elif character == "V":
                self.process_v()
            elif character == "W":
                self.process_w()
            elif character == "X":
                self.process_x()
            elif character == "Z":
                self.process_z()

            # Handle any residual unprocessed sound clusters
            if len(self.next) == 2:
                if self.next[0]:
                    self.primary_phone += self.next[0]
                    self.secondary_phone += self.next[0]
                self.position += self.next[1]
            elif len(self.next) == 3:
                if self.next[0]:
                    self.primary_phone += self.next[0]
                if self.next[1]:
                    self.secondary_phone += self.next[1]
                self.position += self.next[2]

        # Standardize all representations to the dominant
        # first sound clusters
        if len(self.primary_phone) > 4:
            self.primary_phone = self.primary_phone[:4]
        if len(self.secondary_phone) > 4:
            self.secondary_phone = self.secondary_phone[:4]

        if self.primary_phone == self.secondary_phone:
            self.secondary_phone = ""
        return [self.primary_phone, self.secondary_phone]


class Word(object):
    """
    Helper class used for representing a single word for token
    parsing.
    """

    def __init__(self, input):
        self.original = input
        if isinstance(input, bytes):
            self.decoded = input.decode("utf-8", "ignore")
        else:
            self.decoded = input

        # Handle accented characters
        self.decoded = self.decoded.replace("\xc7", "s")
        self.decoded = self.decoded.replace("\xe7", "s")
        self.normalized = "".join(
            (
                c
                for c in unicodedata.normalize("NFD", self.decoded)
                if unicodedata.category(c) != "Mn"
            )
        )
        self.upper = self.normalized.upper()
        self.length = len(self.upper)
        self.prepad = "  "
        self.start_index = len(self.prepad)
        self.end_index = self.start_index + self.length - 1
        self.postpad = "      "

        # Buffer used in algorithm to index outside bounds of original string
        self.buffer = self.prepad + self.upper + self.postpad

    @property
    def is_slavo_germanic(self):
        """
        Use letter clusterings to identify likelihood of language origin.
        """
        return (
            self.upper.find("W") > -1
            or self.upper.find("K") > -1
            or self.upper.find("CZ") > -1
            or self.upper.find("WITZ") > -1
        )

    def get_letters(self, start=0, end=None):
        """
        Return all characters in the requested buffer range.
        """
        if not end:
            end = start + 1
        start = self.start_index + start
        end = self.start_index + end
        return self.buffer[start:end]



from typing import Union

import polling
import requests

from phdi.cloud.core import BaseCredentialManager
from phdi.fhir.transport.http import http_request_with_reauth


def export_from_fhir_server(
    cred_manager: BaseCredentialManager,
    fhir_url: str,
    export_scope: str = "",
    since: str = "",
    resource_type: str = "",
    container: str = "",
    poll_step: float = 30,
    poll_timeout: float = 300,
) -> dict:
    """
    Initiates a FHIR $export operation, polls until it completes, and returns the
    successful result.

    :param cred_manager: The credential manager used to authenticate to the FHIR server.
    :param fhir_url: The FHIR server base URL.
    :param export_scope: Either `Patient` or `Group/[id]` as specified in the FHIR spec
      (https://hl7.org/fhir/uv/bulkdata/export/index.html#bulk-data-kick-off-request).
    :param since: A FHIR instant (https://build.fhir.org/datatypes.html#instant)
      instructing the export to include only resources created or modified after the
      specified instant.
    :param resource_type: A comma-delimited list of FHIR resource types to include
      in exported files.
    :param container: The name of the storage container used to store exported files.
    :param poll_step: The number of seconds to wait between poll requests, waiting
      for export files to be generated.
    :param poll_timeout: The maximum number of seconds to wait for export files to
      be generated.
    :raises polling.TimeoutException: If the FHIR server continually returns a 202
      status indicating in progress until the timeout is reached.
    :return: The JSON-formatted HTTP response of a completed export operation
      as a dictionary.
    """
    # TODO consider putting implementation-specific parameters (e.g. container) in a
    # flexible dictionary rather than listing as explicit parameters.

    # Combine template variables into export endpoint
    access_token = cred_manager.get_access_token()
    export_url = _compose_export_url(
        fhir_url=fhir_url,
        export_scope=export_scope,
        since=since,
        resource_type=resource_type,
        container=container,
    )

    # Open connection to the export operation and kickoff process
    response = http_request_with_reauth(
        cred_manager=cred_manager,
        url=export_url,
        retry_count=3,
        request_type="GET",
        allowed_methods=["GET"],
        headers={
            "Authorization": f"Bearer {access_token}",
            "Accept": "application/fhir+json",
            "Prefer": "respond-async",
        },
    )

    # TODO handle export failure conditions and timeouts.
    if response.status_code == 202:
        # Repeatedly poll the endpoint the FHIR server creates for us
        # until either the connection times out (as we configured) or
        # we have the response in hand
        poll_response = export_from_fhir_server_poll(
            poll_url=response.headers.get("Content-Location"),
            cred_manager=cred_manager,
            poll_step=poll_step,
            poll_timeout=poll_timeout,
        )

        # We successfully completed the full export
        if poll_response.status_code == 200:
            return poll_response.json()


def export_from_fhir_server_poll(
    poll_url: str,
    cred_manager: BaseCredentialManager,
    poll_step: float = 30,
    poll_timeout: float = 300,
) -> requests.Response:
    """
    Polls an endpoint to retrieve an export file after an export run has been initiated.

    :param poll_url: The URL to poll for export information.
    :param cred_manager: The service used to get an access token used to make a request.
    :param poll_step: The number of seconds to wait between poll requests, waiting
      for export files to be generated.
    :param poll_timeout: The maximum number of seconds to wait for export files to
      be generated.
    :raises polling.TimeoutException: If the FHIR server continually returns a 202
      status indicating in progress until the timeout is reached.
    :raises requests.HTTPError: If an unexpected status code is returned.
    :return: A response from polled endpoint.
    """
    response = polling.poll(
        target=_export_from_fhir_server_poll_call,
        args=[poll_url, cred_manager],
        step=poll_step,
        timeout=poll_timeout,
    )

    # If no error conditions, return response
    return response


def _compose_export_url(
    fhir_url: str,
    export_scope: str = "",
    since: str = "",
    resource_type: str = "",
    container: str = "",
) -> str:
    """
    Generates a query string for the export request.  Details can be found
    in the FHIR spec:
    https://hl7.org/fhir/uv/bulkdata/export/index.html#query-parameters

    :param fhir_url: The url of the FHIR server to export from.
    :param export_scope: The data we want back (e.g. Patients).
    :param since: We'll get all FHIR resources that have been updated
      since this given timestamp.
    :param resource_type: A comma-delimited list of resource types we want back.
    :param container: The container where we want to store the uploaded files.
    :return: An export url string.
    """
    export_url = fhir_url
    if export_scope == "Patient" or export_scope.startswith("Group/"):
        export_url += f"/{export_scope}/$export"
    elif export_scope == "":
        export_url += "/$export"
    else:
        raise ValueError("Invalid scope {scope}.  Expected 'Patient' or 'Group/[ID]'.")

    # Start with ? url argument separator, and change it to & after the first parameter
    # is appended to the URL
    separator = "?"
    if since:
        export_url += f"{separator}_since={since}"
        separator = "&"

    if resource_type:
        export_url += f"{separator}_type={resource_type}"
        separator = "&"

    if container:
        export_url += f"{separator}_container={container}"
        separator = "&"

    return export_url


def _export_from_fhir_server_poll_call(
    poll_url: str, cred_manager: BaseCredentialManager
) -> Union[requests.Response, None]:
    """
    Called by `export_from_fhir_server_poll` to see if the export files are ready
    based on received status code. If an export is still in progress, then returns
    `None` so polling continues. If the response is 200, then the export files are
    ready, and returns the HTTP response. Any other status either indicates an error
    or unexpected condition, and an error is raised.

    :param poll_url: The endpoint the FHIR server gave us to query for if
      our files are ready.
    :param cred_manager: The credential manager used to authenticate to the FHIR server.
    :return: An HTTP response (if 200) or None (if still in progress).
    """
    access_token = cred_manager.get_access_token()
    response = http_request_with_reauth(
        cred_manager=cred_manager,
        url=poll_url,
        retry_count=3,
        request_type="GET",
        allowed_methods=["GET"],
        headers={
            "Authorization": f"Bearer {access_token}",
            "Accept": "application/fhir+ndjson",
        },
    )
    if response.status_code == 202:
        # In progress - return None to keep polling
        return
    elif response.status_code == 200:
        # Complete
        return response
    else:
        raise requests.HTTPError(response=response)



from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass
from typing import List
from typing import Optional
from typing import Union


@dataclass
class GeocodeResult:
    """
    Represents a successful geocoding response.
    Based on the field nomenclature of a FHIR address, specified at
    https://www.hl7.org/fhir/datatypes.html#Address.
    """

    line: List[str]
    city: str
    state: str
    postal_code: str
    county_fips: str
    lat: float
    lng: float
    district: Optional[str] = None
    country: Optional[str] = None
    county_name: Optional[str] = None
    precision: Optional[str] = None
    geoid: Optional[str] = None
    census_tract: Optional[str] = None
    census_block: Optional[str] = None


class BaseGeocodeClient(ABC):
    """
    Represents a vendor-agnostic geocoder client. Requires implementing
    classes to define methods to geocode from both strings and dictionaries.
    Callers should use the provided interface functions (e.g., geocode_from_str)
    to interact with the underlying vendor-specific client property.
    """

    @abstractmethod
    def geocode_from_str(self, address: str) -> Union[GeocodeResult, None]:
        """
        Geocodes the provided address, which is formatted as a string.

        :param address: The address to geocode, given as a string.
        :param overwrite: If true, `resource` is modified in-place;
          if false, a copy of `resource` modified and returned.  Default: `True`
        :return: A geocoded address (if valid result) or None (if no valid result).
        """
        pass  # pragma: no cover

    @abstractmethod
    def geocode_from_dict(self, address: dict) -> Union[GeocodeResult, None]:
        """
        Geocodes the provided address, which is formatted as a dictionary.

        The given dictionary should conform to standard nomenclature around address
        fields, including:

        * `street`: the number and street address
        * `street2`: additional street level information (if needed)
        * `apartment`: apartment or suite number (if needed)
        * `city`: city to geocode
        * `state`: state to geocode
        * `postal_code`: the postal code to use
        * `urbanization`: urbanization code for area, sector, or regional
        * `development`: (only used for Puerto Rican addresses)

        There is no minimum number of fields that must be specified to use this
        function; however, a minimum of street, city, and state are suggested
        for the best matches.

        :param address: A dictionary with fields outlined above.
        :return: A geocoded address (if valid result) or None (if no valid result).
        """
        pass  # pragma: no cover



import datetime
import logging
from contextlib import contextmanager
from typing import List

from sqlalchemy import create_engine
from sqlalchemy import MetaData
from sqlalchemy import select
from sqlalchemy import Table
from sqlalchemy import text
from sqlalchemy.orm import scoped_session
from sqlalchemy.orm import sessionmaker


class DataAccessLayer(object):
    """
    Base class for Database API objects - manages transactions,
    sessions and holds a reference to the engine.
    Acts as a simple session context manager and creates a
    uniform API for querying using the ORM
    This class could be thought of as a singleton factory.
    Applications should only ever use one instance per database.
    Example:
        dal = DataAccessLayer()
        dal.connect(engine_url=..., engine_echo=False)
    """

    def __init__(self) -> None:
        self.engine = None
        self.Meta = MetaData()
        self.PATIENT_TABLE = None
        self.PERSON_TABLE = None
        self.NAME_TABLE = None
        self.GIVEN_NAME_TABLE = None
        self.ID_TABLE = None
        self.PHONE_TABLE = None
        self.ADDRESS_TABLE = None
        self.EXTERNAL_PERSON_TABLE = None
        self.EXTERNAL_SOURCE_TABLE = None
        self.TABLE_LIST = []

    def get_connection(
        self,
        engine_url: str,
        engine_echo: bool = False,
        pool_size: int = 5,
        max_overflow: int = 10,
    ) -> None:
        """
        Establish a connection to the database

        this method initiates a connection to the database specified
        by the parameters defined in environment variables. Builds
        engine and Session class for app layer

        :param engine_url: The URL of the database engine
        :param engine_echo: If True, print SQL statements to stdout
        :param pool_size: The number of connections to keep open in the connection pool
        :param max_overflow: The number of connections to allow in the connection pool
          “overflow”
        :return: None
        """

        # create engine/connection
        self.engine = create_engine(
            engine_url,
            client_encoding="utf8",
            echo=engine_echo,
            pool_size=pool_size,
            max_overflow=max_overflow,
        )

    def initialize_schema(self) -> None:
        """
        Initialize the database schema

        This method initializes all the MPI Database tables using SQLAlchemy's
        Table object

        :return: None
        """

        self.PATIENT_TABLE = Table("patient", self.Meta, autoload_with=self.engine)
        self.PERSON_TABLE = Table("person", self.Meta, autoload_with=self.engine)
        self.NAME_TABLE = Table("name", self.Meta, autoload_with=self.engine)
        self.GIVEN_NAME_TABLE = Table(
            "given_name", self.Meta, autoload_with=self.engine
        )
        self.ID_TABLE = Table("identifier", self.Meta, autoload_with=self.engine)
        self.PHONE_TABLE = Table("phone_number", self.Meta, autoload_with=self.engine)
        self.ADDRESS_TABLE = Table("address", self.Meta, autoload_with=self.engine)
        self.EXTERNAL_PERSON_TABLE = Table(
            "external_person", self.Meta, autoload_with=self.engine
        )
        self.EXTERNAL_SOURCE_TABLE = Table(
            "external_source", self.Meta, autoload_with=self.engine
        )

        # order of the list determines the order of
        # inserts due to FK constraints
        self.TABLE_LIST = []
        self.TABLE_LIST.append(self.PERSON_TABLE)
        self.TABLE_LIST.append(self.EXTERNAL_SOURCE_TABLE)
        self.TABLE_LIST.append(self.EXTERNAL_PERSON_TABLE)
        self.TABLE_LIST.append(self.PATIENT_TABLE)
        self.TABLE_LIST.append(self.NAME_TABLE)
        self.TABLE_LIST.append(self.GIVEN_NAME_TABLE)
        self.TABLE_LIST.append(self.ID_TABLE)
        self.TABLE_LIST.append(self.PHONE_TABLE)
        self.TABLE_LIST.append(self.ADDRESS_TABLE)

    @contextmanager
    def transaction(self) -> None:
        """
        Execute a database transaction

        this method safely wraps a session object in a transactional scope
        used for basic create, select, update and delete procedures

        :yield: SQLAlchemy session object
        :raises ValueError: if an error occurs during the transaction
        """
        session = self.get_session()

        try:
            yield session
            session.commit()

        except Exception as error:
            session.rollback()
            raise ValueError(f"{error}")

        finally:
            session.close()

    def bulk_insert_list(
        self, table: Table, records: list[dict], return_primary_keys: bool = True
    ) -> list:
        """
        Perform a bulk insert operation on a table.  A list of records
        as dictionaries are inserted into the specified table.  A list
        of primary keys from the bulk insert can be returned if return_pks
        is set to True.

        :param table_object: the SQLAlchemy table object to insert into
        :param records: a list of records as a dictionaries
        :param return_primary_key: boolean indicating if you want the inserted
            primary keys for the table returned or not, defaults to False
        :return: a list of primary keys or an empty list
        """
        new_primary_keys = []
        if len(records) > 0 and table is not None:
            logging.info(
                f"Getting primary_key_column at:{datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
            )
            primary_key_column = table.primary_key.c[0]
            with self.transaction() as session:
                logging.info(
                    f"Starting session at: {datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
                )
                n_records = 0
                for record in records:
                    n_records += 1
                    if return_primary_keys:
                        logging.info("Returned primary keys")
                        statement = (
                            table.insert().values(record).returning(primary_key_column)
                        )
                        logging.info(
                            f"""Starting statement execution getting
                              new_primary_key for record #{n_records}at:
                                {datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"""  # noqa
                        )
                        new_primary_key = session.execute(statement)
                        # TODO: I don't like this, but seems to
                        # be one of the only ways to get this to work
                        #  I have tried using the column name from the
                        # PK defined in the table and that doesn't work
                        logging.info(
                            f""" Done with statement execution getting new_primary_key
                              for record #{n_records} at: {datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"""  # noqa
                        )
                        new_primary_keys.append(new_primary_key.first()[0])
                    else:
                        logging.info("Did not return primary keys")
                        statement = table.insert().values(record)
                        logging.info(
                            f"Starting statement execution for record #{n_records} at:{datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
                        )
                        session.execute(statement)
                        logging.info(
                            f"""Done with statement execution
                              for record #{n_records} at: {datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"""  # noqa
                        )
        return new_primary_keys

    def bulk_insert_dict(
        self, records_with_table: dict, return_primary_keys: bool = False
    ) -> dict:
        """
        Perform a bulk insert operation on a table as defined
        by the 'table' element in the dictionary along with the
        record(s), a list of record(s) as dictionaries.  This
        allows for several inserts to occur for different tables
        along with a single or multiple records for each table.

        :param records_with_table: a dictionary that defines the
            the SQLAlchemy table name to insert into
            along with a list of dictionaries as records to
            insert into the specified table
            eg. {
            "patient": [{"patient_id": UUID()}],
            "address": [{"line_1": "1313 Mocking Bird Lane, "city": "Scranton"},]
            }
        :param return_primary_keys: boolean indicating if you want the inserted
            primary keys for the table returned or not, defaults to False
        :return: a dictionary that contains table names as keys
            along with a list of the primary keys, if requested.
        """
        return_results = {}
        statements = []
        with self.transaction() as session:
            logging.info(
                f"Starting session at: {datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
            )
            n_records = 0
            for table in self.TABLE_LIST:
                records = records_with_table.get(table.name)
                if records is not None:
                    new_primary_keys = []

                    if len(records) > 0 and table is not None:
                        logging.info(
                            f"Getting primary_key_column at:{datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
                        )
                        primary_key_column = table.primary_key.c[0]

                        for record in records:
                            n_records += 1
                            if return_primary_keys:
                                logging.info("Returned primary keys")
                                statement = (
                                    table.insert()
                                    .values(record)
                                    .returning(primary_key_column)
                                )
                                logging.info(
                                    f"""Starting statement execution getting
                                    new_primary_key for record #{n_records}at:
                                        {datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"""  # noqa
                                )
                                new_primary_key = session.execute(statement)
                                # TODO: I don't like this, but seems to
                                # be one of the only ways to get this to work
                                #  I have tried using the column name from the
                                # PK defined in the table and that doesn't work
                                logging.info(
                                    f""" Done with statement execution getting new_primary_key
                                    for record #{n_records} at: {datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"""  # noqa
                                )
                                new_primary_keys.append(new_primary_key.first()[0])
                            else:
                                logging.info("Did not return primary keys")
                                logging.info(
                                    f"""Starting statement creation for record #{n_records} at:
                                        {datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"""  # noqa
                                )

                                if "dob" in record:
                                    if record["dob"] is not None:
                                        record["dob"] = datetime.datetime.strptime(
                                            record["dob"], "%Y-%m-%d"
                                        )

                                statement = table.insert().values(**record)

                                statement = statement.compile(
                                    self.engine,
                                    compile_kwargs={
                                        "literal_binds": True,
                                        "render_postprocess": str,
                                    },
                                )
                                statement = str(statement)
                                statements.append(statement)
                                logging.info(
                                    f"""Done with statement creation for record #{n_records} at:
                                        {datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"""  # noqa
                                )

                    return_results[table.name] = {"primary_keys": new_primary_keys}

            if not return_primary_keys:
                statements = ";".join(statements)
                logging.info(
                    f"Starting INSERT statement execution at:{datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
                )
                session.execute(text(statements))
                logging.info(
                    f"Done with INSERT statement execution at:{datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
                )
        return return_results

    def select_results(
        self,
        select_statement: select,
        include_col_header: bool = True,
        query_params: dict = None,
    ) -> List[list]:
        """
        Perform a select query and add the results to a
        list of lists.  Then add the column header as the
        first row, in the list of lists if the
        'include_col_header' parameter is True.

        :param select_statement: the select statment to execute
        :param include_col_header: boolean value to indicate if
            one wants to include a top row of the column headers
            or not, defaults to True
        :return: List of lists of select results
        """
        list_results = [[]]
        logging.info(
            f"In select_results, starting new session at {datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
        )
        with self.transaction() as session:
            logging.info(
                f"Starting to execute statement to return results at: {datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
            )
            results = session.execute(select_statement, query_params)

            logging.info(
                f"Done executing statement to return results at: {datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
            )
            list_results = [list(row) for row in results]
            if include_col_header:
                list_results.insert(0, list(results.keys()))
        return list_results

    def get_session(self) -> scoped_session:
        """
        Get a session object

        this method returns a session object to the caller

        :return: SQLAlchemy scoped session
        """

        session = scoped_session(
            sessionmaker(bind=self.engine)
        )  # NOTE extra config can be implemented in this call to sessionmaker factory
        return session()

    def get_table_by_name(self, table_name: str) -> Table:
        """
        Get an SqlAlchemy ORM Table Object based upon the table
        name passed in.

        :param table_name: the name of the table you want to get.
        :return: SqlAlchemy ORM Table Object.
        """

        if table_name is not None and table_name != "":
            # TODO: I am sure there is an easier way to do this
            for table in self.TABLE_LIST:
                if table.name == table_name:
                    return table
        return None

    def get_table_by_column(self, column_name: str) -> Table:
        """
        Finds a table in the MPI based upon the column name.
        Note, this won't work if the column name used exists
        in more than one table.

        :param column_name: the column name you want to find the
            table it belongs to.
        :return: SqlAlchemy ORM Table Object.
        """

        if column_name is not None and column_name != "":
            # TODO: I am sure there is an easier way to do this
            for table in self.TABLE_LIST:
                if column_name in table.c:
                    return table
        return None

    def does_table_have_column(self, table: Table, column_name: str) -> bool:
        """
        Verifies if a column exists in a particular table

        :param table: the table object to verify if column exists
            within.
        :param column_name: the column name you want to verify.
        :return: True or False.
        """
        if table is None or column_name is None or column_name == "":
            return False
        else:
            return column_name in table.c



import xml.etree.ElementTree as et

import hl7
import requests
from lxml import etree

from phdi.cloud.core import BaseCredentialManager
from phdi.fhir.transport import http_request_with_reauth
from phdi.harmonization import standardize_hl7_datetimes
from phdi.transport.http import http_request_with_retry

CCDA_CODES_TO_CONVERSION_RESOURCE = {
    "34133-9": "CCD",
    "11488-4": "ConsultationNote",
    "18842-5": "DischargeSummary",
    "34117-2": "HistoryandPhysical",
    "11504-8": "OperativeNote",
    "28570-0": "ProcedureNote",
    "11506-3": "ProgressNote",
    "57133-1": "ReferralNote",
    "18761-7": "TransferSummary",
}


def add_rr_data_to_eicr(rr, ecr):
    """
    Extracts relevant fields from an RR document, and inserts them into a
    given eICR document. Ensures that the eICR contains properly formatted
    RR fields, including templateId, id, code, title, effectiveTime,
    confidentialityCode, and corresponding entries; and required format tags.

    :param rr: A serialized xml format reportability response (RR) document.
    :param ecr: A serialized xml format electronic initial case report (eICR) document.
    :return: An xml format eICR document with additional fields extracted from the RR.
    """
    # add xmlns:xsi attribute if not there
    lines = ecr.splitlines()
    xsi_tag = "xmlns:xsi"
    if xsi_tag not in lines[0]:
        lines[0] = lines[0].replace(
            'xmlns="urn:hl7-org:v3"',
            'xmlns="urn:hl7-org:v3" '
            'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"',
        )
        ecr = "\n".join(lines)

    rr = etree.fromstring(rr)
    ecr = etree.fromstring(ecr)

    if ecr.xpath('//*[@code="88085-6"]'):
        print("This eCR has already been merged with RR data.")
        return etree.tostring(ecr, encoding="unicode", method="xml")

    # Create the tags for elements we'll be looking for
    rr_tags = [
        "templateId",
        "id",
        "code",
        "title",
        "effectiveTime",
        "confidentialityCode",
    ]
    rr_tags = ["{urn:hl7-org:v3}" + tag for tag in rr_tags]
    rr_elements = []

    # Find root-level elements and add them to a list
    for tag in rr_tags:
        rr_elements.append(rr.find(f"./{tag}", namespaces=rr.nsmap))

    # Find the nested entry element that we need
    entry_tag = "{urn:hl7-org:v3}" + "component/structuredBody/component/section/entry"
    rr_nested_entries = rr.findall(f"./{entry_tag}", namespaces=rr.nsmap)

    organizer_tag = "{urn:hl7-org:v3}" + "organizer"

    # For now we assume there is only one matching entry
    rr_entry = None
    for entry in rr_nested_entries:
        if entry.attrib and "DRIV" in entry.attrib["typeCode"]:
            organizer = entry.find(f"./{organizer_tag}", namespaces=entry.nsmap)
            if (
                organizer is not None
                and "CLUSTER" in organizer.attrib["classCode"]
                and "EVN" in organizer.attrib["moodCode"]
            ):
                rr_entry = entry
                exit

    # find the status in the RR utilizing the templateid root
    # codes specified from the APHL/LAC Spec
    base_tag_for_status = (
        "{urn:hl7-org:v3}" + "component/structuredBody/component/section"
    )
    template_id_tag = "{urn:hl7-org:v3}" + "templateId"
    entry_status_tag = "{urn:hl7-org:v3}" + "entry"
    act_status_tag = "{urn:hl7-org:v3}" + "act"
    sections_for_status = rr.findall(f"./{base_tag_for_status}", namespaces=rr.nsmap)
    rr_entry_for_status_codes = None
    for status_section in sections_for_status:
        template_id = status_section.find(
            f"./{template_id_tag}", namespaces=status_section.nsmap
        )
        if (
            template_id is not None
            and "2.16.840.1.113883.10.20.15.2.2.3" in template_id.attrib["root"]
        ):
            for entry in status_section.findall(
                f"./{entry_status_tag}", namespaces=status_section.nsmap
            ):
                for act in entry.findall(f"./{act_status_tag}", namespaces=entry.nsmap):
                    entry_act_template_id = act.find(
                        f"./{template_id_tag}", namespaces=act.nsmap
                    )
                    if (
                        entry_act_template_id is not None
                        and "2.16.840.1.113883.10.20.15.2.3.29"
                        in entry_act_template_id.attrib["root"]
                    ):
                        # only anticipating one status code
                        rr_entry_for_status_codes = entry
                        exit

    # Create the section element with root-level elements
    # and entry to insert in the eICR
    ecr_section = None
    if rr_entry is not None:
        ecr_section_tag = "{urn:hl7-org:v3}" + "section"
        ecr_section = etree.Element(ecr_section_tag)
        ecr_section.extend(rr_elements)
        if rr_entry_for_status_codes is not None:
            ecr_section.append(rr_entry_for_status_codes)
        ecr_section.append(rr_entry)

        # Append the ecr section into the eCR - puts it at the end
        ecr.append(ecr_section)

    ecr = etree.tostring(ecr, encoding="unicode", method="xml")

    return ecr


def convert_to_fhir(
    message: str,
    url: str,
    cred_manager: BaseCredentialManager = None,
    headers: dict = {},
    use_default_ccda=False,
):
    """
    Converts a given message from either HL7 v2 (pipe-delimited flat file) or CCDA (XML)
    into FHIR format (JSON) for further processing using the FHIR server. Standardizes
    datetimes in HL7v2 messages before conversion.

    This function uses a containerized version of the
    [Azure FHIR Converter](https://github.com/microsoft/FHIR-Converter).

    If conversion succeeds, a `requests.Response` object will be returned with the
    conversion response. Otherwise, a `ConversionError` is raised, with the
    `requests.Response` available as a property for troubleshooting and reporting
    purposes.

    :param message: The raw message that needs to be converted to
      FHIR. Currently, only HL7v2 or CCDA are supported.
    :param url: A URL that points to the location of the converter API.
    :param cred_manager: Service used to get an access token used to
      make a request.
    :param headers: JSON-type dictionary of headers to make the request with.
    :param use_default_ccda: Whether to default to the
      base "CCD" root template if a resource's LOINC code doesn't
      map to a specific supported template (Optional, default is No)
    :raises requests.HttpError: If the HTTP request was unsuccessful.
    :raises ConversionError: If the message could not be converted.
    :return: A requests.Response object

    """
    # TODO Update documentation with a link to the containerized FHIR converter, once
    # it's been ported over to the phdi repository.

    conversion_settings = _get_fhir_conversion_settings(message, use_default_ccda)
    if conversion_settings.get("input_type") == "hl7v2":
        message = standardize_hl7_datetimes(message)

    url = f"{url}"
    data = {
        "input_data": message,
        "input_type": conversion_settings.get("input_type"),
        "root_template": conversion_settings.get("root_template"),
    }

    if cred_manager:
        access_token = cred_manager.get_access_token()
        headers["Authorization"] = f"Bearer {access_token}"
        response = http_request_with_reauth(
            cred_manager=cred_manager,
            url=url,
            retry_count=3,
            request_type="POST",
            allowed_methods=["POST"],
            headers=headers,
            data=data,
        )
    else:
        response = http_request_with_retry(
            url=url,
            retry_count=3,
            request_type="POST",
            allowed_methods=["POST"],
            headers=headers,
            data=data,
        )

    if response.status_code != 200:
        raise ConversionError(response)

    return response


def _get_fhir_conversion_settings(message: str, use_default_ccda=False) -> dict:
    """
    Determines which settings to use with the FHIR server to facilitate message
    conversion by attempting to identify which data type the input has (HL7 or XML)
    and determine the appropriate FHIR converter root template to use. Raises
    an exception if the user opts to not use the default CCDA root template for
    an unsupported input resouece and a message's extracted LOINC code doesn't
    correspond to an existing CCDA template.

    More information about the required templates and settings can be found here:

    https://docs.microsoft.com/en-us/azure/healthcare-apis/azure-api-for-fhir/convert-data

    :param message: The incoming message.
    :param use_default_ccda: Whether to default to the
      base "CCD" root template if a resource's LOINC code doesn't
      map to a specific supported template. Default: `False`
    :raises ConversionError: If conversion settings cannot be derived.
    :return: A dictionary holding the settings of parameters to-be
      set when converting the input to FHIR.
    """
    # Some streams (e.g. ELR, VXU) are HL7v2 encoded
    if message[:3] == "MSH":
        parsed_msg = hl7.parse(message)
        extracted_code = str(parsed_msg.segment("MSH")[9])

        # HL7 MSH segment 9 has three components: message code, trigger
        # event, and message structure. We can extract based on number of
        # present separators and recombine to create a robust formatted code
        extracted_code_tokenized = extracted_code.split(parsed_msg.separators[3])
        formatted_code = ""
        if (len(extracted_code_tokenized) >= 3) and (extracted_code_tokenized[2] != ""):
            formatted_code = extracted_code_tokenized[2]
        elif len(extracted_code_tokenized) == 2:
            formatted_code = (
                f"{extracted_code_tokenized[0]}_{extracted_code_tokenized[1]}"
            )

        if formatted_code == "":
            raise ConversionError(message="Could not determine HL7 message structure")

        return {
            "root_template": formatted_code,
            "input_type": "hl7v2",
        }

    # Others conform to C-CDA standards (e.g. ECR)
    else:
        try:
            root = et.fromstring(message)

            # The Clinical Document tag and codeSystem together denote
            # accepted LOINC codes for convertible resources
            if root.tag.strip() == "{urn:hl7-org:v3}ClinicalDocument":
                for child in root:
                    if (
                        child.tag.strip() == "{urn:hl7-org:v3}code"
                        and child.get("codeSystem") == "2.16.840.1.113883.6.1"
                    ):
                        break
                ccda_code = child.attrib.get("code")

                try:
                    root_template = CCDA_CODES_TO_CONVERSION_RESOURCE[ccda_code]
                    return {
                        "root_template": root_template,
                        "input_type": "ccda",
                    }
                except KeyError:
                    if use_default_ccda:
                        return {
                            "root_template": "CCD",
                            "input_type": "ccda",
                        }
                    else:
                        raise KeyError(
                            "Resource code does not match any provided input template"
                        )

        except et.ParseError as ex:
            raise ConversionError(
                message=(
                    "Input message has unrecognized data type, "
                    + "should be HL7v2 or XML."
                )
            ) from ex


class ConversionError(Exception):
    """
    Exception raised for errors that occur during conversion.
    """

    @property
    def http_response(self):
        return self.__http_response

    def __init__(self, http_response: requests.Response = None, message: str = None):
        """
        Creates a new ConversionError object.

        :param http_response: HTTP response returned by the converter service.
          Default: `None`
        :param message: Error message. Default: `None`
        """
        self.__http_response = http_response

        if (message is None) and http_response is not None:
            message = (
                "Conversion exception occurred with status code"
                + f" {http_response.status_code} returned from the converter service."
            )
        self.message = message

        super().__init__(self.message)



from phdi.linkage.algorithms import DIBBS_BASIC
from phdi.linkage.algorithms import DIBBS_ENHANCED
from phdi.linkage.core import BaseMPIConnectorClient
from phdi.linkage.link import add_person_resource
from phdi.linkage.link import block_data
from phdi.linkage.link import calculate_log_odds
from phdi.linkage.link import calculate_m_probs
from phdi.linkage.link import calculate_u_probs
from phdi.linkage.link import compile_match_lists
from phdi.linkage.link import eval_log_odds_cutoff
from phdi.linkage.link import eval_perfect_match
from phdi.linkage.link import extract_blocking_values_from_record
from phdi.linkage.link import feature_match_exact
from phdi.linkage.link import feature_match_four_char
from phdi.linkage.link import feature_match_fuzzy_string
from phdi.linkage.link import feature_match_log_odds_exact
from phdi.linkage.link import feature_match_log_odds_fuzzy_compare
from phdi.linkage.link import generate_hash_str
from phdi.linkage.link import link_record_against_mpi
from phdi.linkage.link import load_json_probs
from phdi.linkage.link import match_within_block
from phdi.linkage.link import perform_linkage_pass
from phdi.linkage.link import profile_log_odds
from phdi.linkage.link import read_linkage_config
from phdi.linkage.link import score_linkage_vs_truth
from phdi.linkage.link import write_linkage_config
from phdi.linkage.mpi import DIBBsMPIConnectorClient
from phdi.linkage.seed import convert_to_patient_fhir_resources
from phdi.linkage.utils import datetime_to_str

__all__ = [
    "DIBBS_BASIC",
    "DIBBS_ENHANCED",
    "generate_hash_str",
    "block_data",
    "match_within_block",
    "feature_match_exact",
    "feature_match_fuzzy_string",
    "eval_perfect_match",
    "compile_match_lists",
    "feature_match_four_char",
    "perform_linkage_pass",
    "score_linkage_vs_truth",
    "calculate_m_probs",
    "calculate_u_probs",
    "load_json_probs",
    "calculate_log_odds",
    "feature_match_log_odds_exact",
    "feature_match_log_odds_fuzzy_compare",
    "profile_log_odds",
    "eval_log_odds_cutoff",
    "BaseMPIConnectorClient",
    "extract_blocking_values_from_record",
    "write_linkage_config",
    "read_linkage_config",
    "link_record_against_mpi",
    "add_person_resource",
    "_compare_address_elements",
    "_compare_name_elements",
    "convert_to_patient_fhir_resources",
    "DIBBsMPIConnectorClient",
    "datetime_to_str",
]



import json
from typing import List
from typing import Union

import google.auth.transport.requests
from google.auth.credentials import Credentials
from google.cloud import storage

from .core import BaseCloudStorageConnection
from .core import BaseCredentialManager


class GcpCredentialManager(BaseCredentialManager):
    """
    Provides a GCP-specific credential manager.
    """

    @property
    def scope(self) -> str:
        return self.__scope

    @property
    def scoped_credentials(self) -> Credentials:
        return self.__scoped_credentials

    @property
    def project_id(self) -> str:
        return self.__project_id

    def __init__(self, scope: list = None):
        """
        Creates a new GcpCredentialManager object.

        :param scope: A list of scopes to limit access to resource.
        """
        self.__scope = scope
        self.__scoped_credentials = None
        self.__project_id = None

        if self.scope is None:
            self.__scope = ["https://www.googleapis.com/auth/cloud-platform"]

    def get_credential_object(self) -> Credentials:
        """
        Gets a GCP-specific credential object.

        :return: A scoped instance of the Credentials class from google.auth
            package, refreshed if necessary.
        """
        # Get credentials if they don't exist or are expired.
        if self.__scoped_credentials is None:
            self.__scoped_credentials, self.__project_id = google.auth.default(
                scopes=self.scope
            )
        elif self.__scoped_credentials.expired:
            self.__scoped_credentials, self.__project_id = google.auth.default(
                scopes=self.scope
            )

        return self.__scoped_credentials

    def get_project_id(self) -> str:
        """
        Gets the ID of the current GCP project.

        :return: The current GCP project ID.
        """

        if self.__project_id is None:
            self.__scoped_credentials, self.__project_id = google.auth.default(
                scopes=self.scope
            )
        return self.__project_id

    def get_access_token(self) -> str:
        """
        Obtains an access token from GCP.

        :return: The access token, refreshed if necessary.
        """

        creds = self.get_credential_object()
        if not creds.valid:
            request = google.auth.transport.requests.Request()
            creds.refresh(request=request)

        access_token = creds.token

        return access_token


class GcpCloudStorageConnection(BaseCloudStorageConnection):
    """
    Defines a connection used for interacting with cloud storage in GCP.
    """

    @property
    def storage_client(self) -> storage.Client:
        return self.__storage_client

    def __init__(self):
        """
        Creates a new GcpCloudContainerConnection object.
        """

        self.__storage_client = None

    def _get_storage_client(self) -> storage.Client:
        """
        Obtains a client connected to an GCP storage container by
        utilizing the first valid credentials GCP can find. Credential validation
        is handeled by GCP.

        :return: The GCP storage client.
        """

        if self.__storage_client is None:
            self.__storage_client = storage.Client()

        return self.__storage_client

    def download_object(
        self, container_name: str, filename: str, encoding: str = "utf-8"
    ) -> str:
        """
        Downloads a character blob from storage and returns it as a string.

        :param container_name: The name of the bucket containing object to download.
        :param filename: The location of file within GCP blob storage.
        :param encoding: The encoding applied to the downloaded content.
        :return: A character blob (as a string) from given bucket and filename.
        """
        storage_client = self._get_storage_client()
        blob = storage_client.bucket(container_name).blob(filename)
        return blob.download_as_text(encoding=encoding)

    def upload_object(
        self,
        message: Union[str, dict],
        container_name: str,
        filename: str,
        content_type="application/json",
    ) -> None:
        """
        Uploads the content of a given message to GCP blob storage.
        The message can be passed either as a raw string or as JSON.

        :param message: The contents of a message, encoded either as a
          string or in a JSON-formatted dictionary.
        :param container_name: The name of the target bucket for upload.
        :param filename: The location of file within GCP blob storage.
        """

        storage_client = self._get_storage_client()
        bucket = storage_client.bucket(container_name)

        blob = bucket.blob(filename)

        if isinstance(message, dict):
            message = json.dumps(message).encode("utf-8")

        blob.upload_from_string(data=message, content_type=content_type)

    def list_containers(self) -> List[str]:
        """
        Lists bucket names in storage.

        :return: A list of bucket names in storage.
        """
        storage_client = self._get_storage_client()

        bucket_name_list = []
        for bucket in storage_client.list_buckets():
            bucket_name_list.append(bucket)
        return bucket_name_list

    def list_objects(self, container_name: str, prefix: str = "") -> List[str]:
        """
        Lists names for objects within a bucket.

        :param container_name: The name of the bucket to look for objects.
        :param prefix: Filter the objects returned to filenames beginning
          with this value.
        :return: A list of names for objects in given bucket.
        """
        storage_client = self._get_storage_client()

        blob_properties_generator = storage_client.list_blobs(
            container_name, prefix=prefix
        )

        blob_name_list = []
        for blob_propreties in blob_properties_generator:
            blob_name_list.append(blob_propreties.name)

        return blob_name_list



from phdi.geospatial.census import CensusGeocodeClient
from phdi.geospatial.core import BaseGeocodeClient
from phdi.geospatial.core import GeocodeResult
from phdi.geospatial.smarty import SmartyGeocodeClient

__all__ = (
    "GeocodeResult",
    "BaseGeocodeClient",
    "SmartyGeocodeClient",
    "CensusGeocodeClient",
)



# Pre-computed log-odds points values for each of the DIBBs MPI
# supported columns (derived from representative synthetic data)
LOG_ODDS_SCORES = {
    "birthdate": 9.944142836217619,
    "first_name": 8.009121400325398,
    "last_name": 5.327681398982514,
    "sex": 0.6964525713514773,
    "address": 5.769942276960749,
    "city": 1.8002552875091014,
    "state": 0.0,
    "zip": 4.909466232098861,
    "mrn": 1.464232660081324,
}


DIBBS_BASIC = [
    {
        "funcs": {
            "first_name": "feature_match_fuzzy_string",
            "last_name": "feature_match_fuzzy_string",
            "birthdate": "feature_match_fuzzy_string",
        },
        "blocks": [
            {"value": "mrn", "transformation": "last4"},
            {"value": "address", "transformation": "first4"},
        ],
        "matching_rule": "eval_perfect_match",
        "cluster_ratio": 0.9,
    },
    {
        "funcs": {
            "address": "feature_match_fuzzy_string",
            "city": "feature_match_fuzzy_string",
        },
        "blocks": [
            {"value": "first_name", "transformation": "first4"},
            {"value": "last_name", "transformation": "first4"},
        ],
        "matching_rule": "eval_perfect_match",
        "cluster_ratio": 0.9,
    },
]


DIBBS_ENHANCED = [
    {
        "funcs": {
            "birthdate": "feature_match_log_odds_fuzzy_compare",
            "first_name": "feature_match_log_odds_fuzzy_compare",
            "last_name": "feature_match_log_odds_fuzzy_compare",
        },
        "blocks": [
            {"value": "mrn", "transformation": "last4"},
            {"value": "address", "transformation": "first4"},
        ],
        "matching_rule": "eval_log_odds_cutoff",
        "cluster_ratio": 0.9,
        "kwargs": {
            "similarity_measure": "JaroWinkler",
            "threshold": 0.7,
            "true_match_threshold": 16.5,
            "log_odds": LOG_ODDS_SCORES,
        },
    },
    {
        "funcs": {
            "address": "feature_match_log_odds_fuzzy_compare",
            "city": "feature_match_log_odds_fuzzy_compare",
        },
        "blocks": [
            {"value": "first_name", "transformation": "first4"},
            {"value": "last_name", "transformation": "first4"},
        ],
        "matching_rule": "eval_log_odds_cutoff",
        "cluster_ratio": 0.9,
        "kwargs": {
            "similarity_measure": "JaroWinkler",
            "threshold": 0.7,
            "true_match_threshold": 7.0,
            "log_odds": LOG_ODDS_SCORES,
        },
    },
]



from abc import ABC
from abc import abstractmethod
from typing import List

from sqlalchemy import Select


class BaseMPIConnectorClient(ABC):
    """
    Represents a vendor-agnostic Master Patient Index (MPI) connector client. Requires
    implementing classes to define methods to retrive blocks of data from the MPI.
    Callers should use the provided interface functions (e.g., block_data)
    to interact with the underlying vendor-specific client property.
    """

    @abstractmethod
    def get_block_data() -> List[list]:
        """
        Returns a list of lists containing records from the MPI database that
        match on the incoming record's block criteria and values. If blocking
        on 'ZIP' and the incoming record's zip code is '90210', the resulting
        block of data would contain records that all
        have the same zip code of 90210.

        """
        pass  # pragma: no cover

    @abstractmethod
    def insert_matched_patient() -> None:
        """
        If a matching person ID has been found in the MPI, inserts a new patient into
        the patient table and all other subsequent MPI tables, including the
        matched person id, to link the new patient and matched person ID;
        else inserts a new patient into the patient table, as well as all other
        subsequent MPI tables, and inserts a new person into the person table
        linking the new person to the new patient.
        """
        pass  # pragma: no cover

    @abstractmethod
    def _generate_block_query(
        self, organized_block_criteria: dict, query: Select
    ) -> Select:
        """
         Generates a query for selecting a block of data from the MPI tables per the
        block field criteria.  The block field criteria should be a dictionary
        organized by MPI table name, with the ORM table object, and the blocking
        criteria.
        """
        pass  # pragma: no cover



from phdi.validation.validation import _add_message_ids
from phdi.validation.validation import _append_error_message
from phdi.validation.validation import _clear_all_errors_and_ids
from phdi.validation.validation import _organize_error_messages
from phdi.validation.validation import _response_builder
from phdi.validation.validation import validate_ecr
from phdi.validation.xml_utils import _check_xml_names_and_attribs_exist
from phdi.validation.xml_utils import _get_ecr_custom_message
from phdi.validation.xml_utils import _get_xml_attributes
from phdi.validation.xml_utils import _get_xml_message_id
from phdi.validation.xml_utils import _get_xml_relative_iterator
from phdi.validation.xml_utils import _get_xml_relatives_details
from phdi.validation.xml_utils import _validate_xml_related_element
from phdi.validation.xml_utils import _validate_xml_relatives
from phdi.validation.xml_utils import ECR_NAMESPACES
from phdi.validation.xml_utils import get_ecr_message_ids
from phdi.validation.xml_utils import get_xml_element_details
from phdi.validation.xml_utils import validate_xml_attributes
from phdi.validation.xml_utils import validate_xml_elements
from phdi.validation.xml_utils import validate_xml_value

__all__ = [
    "validate_ecr",
    "_organize_error_messages",
    "_response_builder",
    "_append_error_message",
    "_add_message_ids",
    "_clear_all_errors_and_ids",
    "get_ecr_message_ids",
    "_check_xml_names_and_attribs_exist",
    "_get_ecr_custom_message",
    "_get_xml_message_id",
    "get_xml_element_details",
    "_get_xml_attributes",
    "_get_xml_relative_iterator",
    "validate_xml_elements",
    "validate_xml_attributes",
    "_validate_xml_related_element",
    "_validate_xml_relatives",
    "validate_xml_value",
    "_get_xml_relatives_details",
    "ECR_NAMESPACES",
]



from importlib import metadata
from pathlib import Path
from typing import Literal

from fastapi import FastAPI
from pydantic import BaseModel

# create a class with the DIBBs default Creative Commons Zero v1.0 and
# MIT license to be used by the BaseService class
LICENSES = {
    "CreativeCommonsZero": {
        "name": "Creative Commons Zero v1.0 Universal",
        "url": "https://creativecommons.org/publicdomain/zero/1.0/",
    },
    "MIT": {"name": "The MIT License", "url": "https://mit-license.org/"},
}

DIBBS_CONTACT = {
    "name": "CDC Public Health Data Infrastructure",
    "url": "https://cdcgov.github.io/phdi-site/",
    "email": "dmibuildingblocks@cdc.gov",
}


STATUS_OK = {"status": "OK"}


class StatusResponse(BaseModel):
    """
    The schema for the response from the health check endpoint.
    """

    status: Literal["OK"]


class BaseService:
    def __init__(
        self,
        service_name: str,
        service_path: str,
        description_path: str,
        include_health_check_endpoint: bool = True,
        license_info: Literal["CreativeCommonsZero", "MIT"] = "CreativeCommonsZero",
    ):
        """
        Initialize a BaseService instance.

        :param service_name: The name of the service.
        :param service_path: The path to used to access the service from a gateway.
        :param description_path: The path to a markdown file containing a description of
            the service.
        :param include_health_check_endpoint: If True, the standard DIBBs health check
            endpoint will be added.
        :param license_info: If empty, the standard DIBBs Creative Commons Zero v1.0
            Universal license will be used. The other available option is to use the
            MIT license.
        """
        description = Path(description_path).read_text(encoding="utf-8")
        self.service_path = service_path
        self.include_health_check_endpoint = include_health_check_endpoint
        self.app = FastAPI(
            title=service_name,
            version=metadata.version("phdi"),
            contact=DIBBS_CONTACT,
            license_info=LICENSES[license_info],
            description=description,
        )

    def add_path_rewrite_middleware(self):
        """
        Add middleware to the FastAPI instance to strip the service_path
        from the URL path if it is present. This is useful when the service
        is behind a gateway that is using a path-based routing strategy.
        """

        @self.app.middleware("http")
        async def rewrite_path(request, call_next):
            if request.url.path.startswith(self.service_path):
                request.scope["path"] = request.scope["path"].replace(
                    self.service_path, ""
                )
                if request.scope["path"] == "":
                    request.scope["path"] = "/"
            return await call_next(request)

    def add_health_check_endpoint(self):
        @self.app.get("/")
        async def health_check() -> StatusResponse:
            """
            Check service status. If an HTTP 200 status code is returned along with
            '{"status": "OK"}' then the service is available and running properly.
            """
            return STATUS_OK

    def start(self) -> FastAPI:
        """
        Return a FastAPI instance with DIBBs metadata set. If
        `include_health_check_endpoint` is True, then the health check endpoint
        will be added.

        :return: The FastAPI instance.
        """
        self.add_path_rewrite_middleware()
        if self.include_health_check_endpoint:
            self.add_health_check_endpoint()
        return self.app



import copy

from smartystreets_python_sdk import us_street

from phdi.fhir.geospatial.core import BaseFhirGeocodeClient
from phdi.fhir.utils import get_one_line_address
from phdi.geospatial.smarty import SmartyGeocodeClient


class SmartyFhirGeocodeClient(BaseFhirGeocodeClient):
    """
    Implementation of a geocoding client designed to handle FHIR-
    formatted data using the SmartyStreets API.
    Requires an authorization ID as well as an authentication token
    in order to build a street lookup client.
    """

    def __init__(
        self,
        smarty_auth_id: str,
        smarty_auth_token: str,
        licenses: list[str] = ["us-standard-cloud"],
    ):
        self.__client = SmartyGeocodeClient(smarty_auth_id, smarty_auth_token, licenses)

    @property
    def geocode_client(self) -> us_street.Client:
        """
        An instance of the underlying Smarty client.
        Allows the FHIR wrapper to access a SmartyStreets-
        specific connection client without instantiating its own
        copy. Provides access to the respective `geocode_from_str`
        and `geocode_from_dict` methods if they're desired.
        """
        return self.__client

    def geocode_resource(self, resource: dict, overwrite=True) -> dict:
        """
        Performs geocoding on one or more addresses in a given FHIR
        resource and returns either the result or a copy thereof.
        Currently supported resource types are:

        * Patient

        :param resource: The resource whose addresses should be geocoded.
        :param overwrite: If true, `resource` is modified in-place;
          if false, a copy of `resource` modified and returned.  Default: `True`
        :return: The geocoded resource as a dict.
        """
        if not overwrite:
            resource = copy.deepcopy(resource)

        resource_type = resource.get("resourceType", "")
        if resource_type == "Patient":
            self._geocode_patient_resource(resource)

        return resource

    def _geocode_patient_resource(self, patient: dict) -> None:
        """
        Geocodes all addresses in a patient resource.

        :param patient: A FHIR Patient resource.
        """
        for address in patient.get("address", []):
            address_str = get_one_line_address(address)
            standardized_address = self.__client.geocode_from_str(address_str)

            # Update fields with new, standardized information
            if standardized_address:
                address["line"] = standardized_address.line
                address["city"] = standardized_address.city
                address["state"] = standardized_address.state
                address["county"] = standardized_address.county_name
                address["postalCode"] = standardized_address.postal_code
                self._store_lat_long_extension(
                    address, standardized_address.lat, standardized_address.lng
                )

    def geocode_bundle(self, bundle: dict, overwrite=True) -> dict:
        """
        Geocodes on all resources in a given FHIR bundle whose
        resource type is among those supported by the PHDI SDK. Currently,
        this includes:

        * Patient

        :param bundle: A bundle of FHIR resources.
        :param overwrite: If true, `bundle` is modified in-place;
          if false, a copy of `bundle` modified and returned.  Default: `True`
        :return: The FHIR bundle with geocoded address(es).
        """
        if not overwrite:
            bundle = copy.deepcopy(bundle)

        for entry in bundle.get("entry", []):
            _ = self.geocode_resource(entry.get("resource", {}), overwrite=True)

        return bundle



import datetime
import logging
import uuid
from functools import cache
from typing import Dict
from typing import List
from typing import Literal

from sqlalchemy import and_
from sqlalchemy import Select
from sqlalchemy import select
from sqlalchemy import text
from sqlalchemy.dialects.postgresql import aggregate_order_by
from sqlalchemy.dialects.postgresql import array_agg

from phdi.fhir.utils import extract_value_with_resource_path
from phdi.linkage.core import BaseMPIConnectorClient
from phdi.linkage.dal import DataAccessLayer
from phdi.linkage.utils import load_mpi_env_vars_os


class DIBBsMPIConnectorClient(BaseMPIConnectorClient):
    """
    Represents a Postgres-specific Master Patient Index (MPI) connector
    client for the DIBBs implementation of the record linkage building
    block. Callers should use the provided interface functions (e.g.,
    block_vals) to interact with the underlying vendor-specific client
    property.

    """

    def __init__(self, pool_size: int = 5, max_overflow: int = 10):
        """
        Initialize the MPI connector client with the MPI database.
        :param pool_size: The number of connections to keep open to the database.
        :param max_overflow: The number of connections to allow in connection pool.
        """
        dbsettings = load_mpi_env_vars_os()
        dbuser = dbsettings.get("user")
        dbname = dbsettings.get("dbname")
        dbpwd = dbsettings.get("password")
        dbhost = dbsettings.get("host")
        dbport = dbsettings.get("port")
        self.dal = DataAccessLayer()
        self.dal.get_connection(
            engine_url=f"postgresql+psycopg2://{dbuser}:"
            + f"{dbpwd}@{dbhost}:{dbport}/{dbname}",
            pool_size=pool_size,
            max_overflow=max_overflow,
        )
        self.dal.initialize_schema()
        self.column_to_fhirpaths = {
            "patient": {
                "root_path": "Patient",
                "fields": {
                    "patient_id": "id",
                    "person_id": "person",
                    "dob": "birthDate",
                    "sex": "gender",
                    "race": """
                        extension.where(url =
                         'http://hl7.org/fhir/us/core/StructureDefinition/us-core-race')
                        .extension.valueCoding.display""",
                    "ethnicity": """
                    extension.where(url =
                    'http://hl7.org/fhir/us/core/StructureDefinition/us-core-ethnicity')
                    .extension.valueCoding.display""",
                },
            },
            "name": {
                "root_path": "Patient.name",
                "fields": {
                    "last_name": "family",
                    "given_name": "given",
                    "type": "use",
                },
            },
            "phone_number": {
                "root_path": "Patient.telecom.where(system='phone')",
                "fields": {
                    "phone_number": "value",
                    "start_date": "period.start",
                    "end_date": "period.end",
                    "type": "use",
                },
            },
            "address": {
                "root_path": "Patient.address",
                "fields": {
                    "line_1": "line[0]",
                    "line_2": "line[1]",
                    "city": "city",
                    "state": "state",
                    "zip_code": "postalCode",
                    "country": "country",
                    "latitude": """
                        extension.where(url =
                          'http://hl7.org/fhir/StructureDefinition/geolocation')
                        .extension.where(url = 'latitude').valueDecimal""",
                    "longitude": """extension.where(url =
                          'http://hl7.org/fhir/StructureDefinition/geolocation')
                          .extension.where(url = 'longitude').valueDecimal""",
                    "type": "use",
                    "start_date": "period.start",
                    "end_date": "period.end",
                },
            },
            "identifier": {
                "root_path": "Patient.identifier",
                "fields": {
                    "patient_identifier": "value",
                    "type_code": "type.coding[0].code",
                    "type_display": "type.coding[0].display",
                    "type_system": "type.coding[0].system",
                },
            },
        }

    def get_block_data(self, block_criteria: Dict) -> List[list]:
        """
        Returns a list of lists containing records from the MPI database that
        match on the incoming record's block criteria and values. If blocking
        on 'ZIP' and the incoming record's ip code is '90210', the resulting
        block of data would contain records that all
        have the same zip code of 90210.

        :param block_criteria: Dictionary containing key value pairs
            for the column name for blocking and the data for the
            incoming record as well as any transformations,
          e.g., {"ZIP": {"value": "90210"}} or
          {"ZIP": {"value": "90210",}, "transformation":"first4"}.
        :return: A list of records that are within the block, e.g.,
            records that all have 90210 as their ZIP.
        """
        logging.info("In get_block_data")
        if len(block_criteria) == 0:
            raise ValueError("`block_vals` cannot be empty.")

        # Get the base query that will select all necessary
        # columns for linkage with some basic filtering
        logging.info(
            f"Starting _get_base_query at:{datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
        )
        query = self._get_base_query()
        logging.info(
            f"Done with _get_base_query at:{datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
        )

        # now get the criteria organized by table so the
        # CTE queries can be constructed and then added
        # to the base query
        logging.info(
            f"Starting _organize_block_criteria at:{datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
        )
        organized_block_vals = self._organize_block_criteria(block_criteria)
        logging.info(
            f"Done with _organize_block_criteria at:{datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}",  # noqa
        )

        # now tack on the where criteria using the block_vals
        # while ensuring they exist in the table structure ORM
        logging.info(
            f"Starting _generate_block_query at:{datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
        )
        query_w_ctes, query_params = self._generate_block_query(
            organized_block_criteria=organized_block_vals, query=query
        )
        logging.info(
            f"Done with _generate_block_query at:{datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
        )
        logging.info(
            f"Starting dal.select_results at:{datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
        )
        blocked_data = self.dal.select_results(
            select_statement=query_w_ctes,
            query_params=query_params,
            include_col_header=True,
        )
        logging.info(
            f"Done with dal.select_results at:{datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
        )

        return blocked_data

    def insert_matched_patient(
        self,
        patient_resource: Dict,
        person_id=None,
        external_person_id=None,
    ) -> str:
        """
        If a matching person ID has been found in the MPI, inserts a new patient into
        the patient table and all other subsequent MPI tables, including the
        matched person id, to link the new patient and matched person ID;
        else inserts a new patient into the patient table, as well as all other
        subsequent MPI tables, and inserts a new person into the person table
        linking the new person to the new patient.

        :param patient_resource: A FHIR patient resource.
        :param person_id: The person ID matching the patient record if a match has been
          found in the MPI, defaults to None.
        :param external_person_id: The external person id for the person that matches
          the patient record if a match has been found in the MPI, defaults to None.
        """
        logging.info(
            f"Starting insert_matched_patient at {datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
        )

        try:
            if person_id is None:
                logging.info(
                    f"person_id was None; starting _insert_person at: {datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
                )
                person_id = self._insert_person()
                logging.info(
                    f"person_id was None; done with _insert_person at: {datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
                )
            patient_resource["person"] = person_id
            logging.info(
                f"Starting _get_mpi_records at: {datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
            )
            mpi_records = self._get_mpi_records(patient_resource)
            logging.info(
                f"Done with _get_mpi_records at:{datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
            )
            logging.info(
                f"Starting dal.bulk_insert_dict at: {datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
            )
            self.dal.bulk_insert_dict(
                records_with_table=mpi_records, return_primary_keys=False
            )
            logging.info(
                f"Done with dal.bulk_insert_dict at:{datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
            )

            if external_person_id is not None:
                logging.info(
                    f"""external_person_id was not None;
                      starting _insert_external_person_id at:{datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"""  # noqa
                )
                self._insert_external_person_id(person_id, external_person_id)
                logging.info(
                    f"""external_person_id was not None;
                      done with _insert_external_person_id at:{datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"""  # noqa
                )
        except Exception as error:  # pragma: no cover
            raise ValueError(f"{error}")

        return person_id

    def _generate_where_clause(
        self, block_criteria: dict, table_name: str
    ) -> tuple[str, list]:
        """
        Generates a list of where criteria leveraging the blocking criteria,
        including transformations such as 'first 4' or 'last 4'.  This
        function leverages the ORM to determine the table and columns.

        :param block_criteria: a dictionary that contains the blocking criteria.
        :return: A tuple containing the query's where clause as a string and a list of
        the appropriate params to pass into the where clause.

        """
        where_placeholders = []
        where_params = {}

        for key, value in block_criteria.items():
            placeholder = f"criterion_{table_name}_{key}"
            criteria_value = value["value"]
            criteria_transform = value.get("transformation", None)

            if criteria_transform is None:
                where_placeholders.append(f"{table_name}.{key} = :{placeholder}")
            else:
                if criteria_transform == "first4":
                    where_placeholders.append(
                        f"LEFT({table_name}.{key},4) = :{placeholder}"
                    )
                elif criteria_transform == "last4":
                    where_placeholders.append(
                        f"RIGHT({table_name}.{key},4) = :{placeholder}"
                    )
            where_params[placeholder] = criteria_value

        where_clause = " AND ".join(where_placeholders)

        return where_clause, where_params

    def _generate_block_query(
        self, organized_block_criteria: dict, query: Select
    ) -> tuple[Select, dict]:
        """
        Generates a query for selecting a block of data from the MPI tables per the
        block field criteria.  The block field criteria should be a dictionary
        organized by MPI table name, with the ORM table object, and the blocking
        criteria.

        :param organized_block_vals: a dictionary organized by MPI table name,
            with the ORM table object, and the blocking criteria.
        :return: A tuple of a 'Select' statement built by the sqlalchemy ORM utilizing
            the blocking criteria and a dictionary of the query parameters.

        """
        new_query = query
        new_query_params = {}

        for table_key, table_info in organized_block_criteria.items():
            where_params = None
            cte_query = None
            sub_query = None

            cte_query_table = table_info["table"]
            where_clause, where_params = self._generate_where_clause(
                table_info["criteria"], table_key
            )
            new_query_params = {**new_query_params, **where_params}

            if where_clause != "" and len(where_params) > 0:
                if self.dal.does_table_have_column(cte_query_table, "patient_id"):
                    cte_query = (
                        select(cte_query_table.c.patient_id.label("patient_id"))
                        .distinct()
                        .where(text(where_clause))
                        .cte(f"{table_key}_cte")
                    )
                else:
                    fk_info = next(iter(cte_query_table.foreign_keys))
                    fk_column = fk_info.column
                    fk_table = fk_info.column.table
                    sub_query = (
                        select(cte_query_table)
                        .where(text(where_clause))
                        .subquery(f"{cte_query_table.name}_cte_subq")
                    )
                    cte_query = (
                        select(fk_table.c.patient_id)
                        .distinct()
                        .join(
                            sub_query,
                            text(
                                f"{fk_table.name}.{fk_column.name} = "
                                + f"{sub_query.name}.{fk_column.name}"
                            ),
                        )
                    ).cte(f"{table_key}_cte")
            if cte_query is not None:
                new_query = new_query.join(
                    cte_query,
                    and_(cte_query.c.patient_id == self.dal.PATIENT_TABLE.c.patient_id),
                )

        return new_query, new_query_params

    def _organize_block_criteria(self, block_fields: dict) -> dict:
        """
        Creates a dictionary with the MPI Table Names as keys
        that also stores the ORM Table Object, for each table,
        as well as the blocking criteria for each of the tables.
        The Table is discovered based upon the blocking columns.
        There is some transformation that occurs between certain
        blocking column criteria and the actual column names.
        Accepted blocking fields include: first_name, last_name,
        birthdate, address line 1, city, state, zip, mrn, and sex.

        :param block_fields: A dictionary that contains the
            configured block fields along with the criteria
            and values for blocking.
        :return: A dictionary organized by table name that
            has the ORM Table Object and all blocking
            criteria and values.
        """
        # Accepted blocking fields include: first_name, last_name,
        # birthdate, address line 1, city, state, zip, mrn, and sex.
        organized_block_vals = {}
        for block_key, block_value in block_fields.items():
            sub_dict = {}
            # TODO: we may find a better way to handle this, but for now
            # just convert the known fields into their proper column counterparts
            if block_key == "address":
                sub_dict["line_1"] = block_value
                table_orm = self.dal.get_table_by_column("line_1")
            elif block_key == "zip":
                sub_dict["zip_code"] = block_value
                table_orm = self.dal.get_table_by_column("zip_code")
            elif block_key == "first_name":
                sub_dict["given_name"] = block_value
                table_orm = self.dal.get_table_by_column("given_name")
            elif block_key == "birthdate":
                sub_dict["dob"] = block_value
                table_orm = self.dal.get_table_by_column("dob")
            elif block_key == "mrn":
                sub_dict["patient_identifier"] = block_value
                # mrn specific criteria
                sub_dict["type_code"] = {"value": "MR"}
                table_orm = self.dal.get_table_by_column("patient_identifier")
            else:
                sub_dict[block_key] = block_value
                table_orm = self.dal.get_table_by_column(block_key)
            if table_orm is not None:
                if table_orm.name in organized_block_vals.keys():
                    organized_block_vals[table_orm.name]["criteria"].update(sub_dict)
                else:
                    organized_block_vals[table_orm.name] = {
                        "table": table_orm,
                        "criteria": sub_dict,
                    }
            else:
                continue
        return organized_block_vals

    def _get_base_query(self) -> Select:
        """
        Generates a select query that pulls all the relevant
        MPI records from the MPI tables, using an ORM, for
        Patient Matching/Blocking.

        :return: A single select statement queries all relevant
            blocking columns and tables from the MPI.
        """

        id_sub_query = (
            select(
                self.dal.ID_TABLE.c.patient_identifier.label("mrn"),
                self.dal.ID_TABLE.c.patient_id.label("patient_id"),
            )
            .where(self.dal.ID_TABLE.c.type_code == "MR")
            .subquery("ident_subq")
        )

        # TODO: keeping this here for the time
        # when we decide to add phone numbers into
        # the blocking data
        #
        #  phone_sub_query = (
        #     select(
        #         self.dal.PHONE_TABLE.c.phone_number.label("phone_number"),
        #         self.dal.PHONE_TABLE.c.type.label("phone_type"),
        #         self.dal.PHONE_TABLE.c.patient_id.label("patient_id"),
        #     )
        #     .where(self.dal.PHONE_TABLE.c.type.in_(["home", "cell"]))
        #     .subquery()
        # )

        query = (
            select(
                self.dal.PATIENT_TABLE.c.patient_id,
                self.dal.PATIENT_TABLE.c.person_id,
                self.dal.PATIENT_TABLE.c.dob.label("birthdate"),
                self.dal.PATIENT_TABLE.c.sex,
                id_sub_query.c.mrn,
                self.dal.NAME_TABLE.c.last_name,
                # Aggregate the given names into an array ordered by the name index
                array_agg(
                    aggregate_order_by(
                        self.dal.GIVEN_NAME_TABLE.c.given_name,
                        self.dal.GIVEN_NAME_TABLE.c.given_name_index.asc(),
                    )
                ).label("given_name"),
                # TODO: keeping this here for the time
                # when we decide to add phone numbers into
                # the blocking data
                #
                # phone_sub_query.c.phone_number,
                # phone_sub_query.c.phone_type,
                self.dal.ADDRESS_TABLE.c.line_1.label("address"),
                self.dal.ADDRESS_TABLE.c.zip_code.label("zip"),
                self.dal.ADDRESS_TABLE.c.city,
                self.dal.ADDRESS_TABLE.c.state,
            )
            .outerjoin(
                id_sub_query,
            )
            .outerjoin(self.dal.NAME_TABLE)
            .outerjoin(self.dal.GIVEN_NAME_TABLE)
            # TODO: keeping this here for the time
            # when we decide to add phone numbers into
            # the blocking data
            #
            # .outerjoin(phone_sub_query)
            .outerjoin(self.dal.ADDRESS_TABLE)
            .group_by(
                self.dal.PATIENT_TABLE.c.patient_id,
                self.dal.PATIENT_TABLE.c.person_id,
                "birthdate",
                self.dal.PATIENT_TABLE.c.sex,
                id_sub_query.c.mrn,
                self.dal.NAME_TABLE.c.last_name,
                self.dal.NAME_TABLE.c.name_id,
                "address",
                "zip",
                self.dal.ADDRESS_TABLE.c.city,
                self.dal.ADDRESS_TABLE.c.state,
            )
        )
        return query

    def _get_mpi_records(self, patient_resource: dict) -> dict:
        """
        Generates a dictionary with the different MPI Table
        Name as keys along with the records for each of the
        MPI Tables, based upon the FHIR Patient Resource data
        passed in.
        There are cases where a direct insert occurs to get
        the primary key, that will be used as a foreign key
        in another MPI Table record.  ie. patient_id is used
        in almost every table, so a patient insert must occur
        first to get the Patient primary key for the other
        MPI table foreign keys.


        :param patient_resource: The FHIR Patient Resource that
            contains patient data to create new MPI records.
        :return: A dictionary of MPI Table names and records.
        """
        records = {}
        if patient_resource["resourceType"] != "Patient":
            return records

        # Check if patient_id exists
        if patient_resource.get("id", None) is None:
            patient_id = uuid.uuid4()
            patient_resource["id"] = patient_id
        else:
            patient_id = patient_resource.get("id")

        for table in self.column_to_fhirpaths.keys():
            table_dict = self.column_to_fhirpaths.get(table)
            table_fields = table_dict.get("fields")

            # Parse root path
            root = extract_value_with_resource_path(
                patient_resource, table_dict.get("root_path"), selection_criteria="all"
            )
            # Parse fields
            table_records = []
            if root is not None:
                for element in root:
                    record = {"patient_id": patient_id}
                    if table == "name":
                        name_id = uuid.uuid4()
                        record["name_id"] = name_id
                    for field in table_fields.keys():
                        selection_criteria = "first"
                        if field == "given_name":
                            selection_criteria = "all"

                        value = extract_value_with_resource_path(
                            element,
                            table_fields.get(field),
                            selection_criteria=selection_criteria,
                        )
                        # Create given_name table in records
                        if field == "given_name":
                            given_name_table_records = self._extract_given_names(
                                value, name_id
                            )
                            if field not in records.keys():
                                records[field] = given_name_table_records
                            else:
                                for given_name_table_record in given_name_table_records:
                                    records[field].append(given_name_table_record)
                            continue
                        record[field] = value

                    table_records.append(record)

            records[table] = table_records

        return records

    def _extract_given_names(self, given_names: list, name_id: uuid) -> dict:
        """
        Separates given_name into it's own table and creates the given_name_index
        ahead of inserting the table into the MPI.

        :param given_names: List of given names.
        :return: List of dictionaries containing entries for the given_name table with
            1 given name and index per row as well as an associated given_name_id.
        """
        table_records = []
        if given_names is not None:
            for idx, name in enumerate(given_names):
                record = {
                    "name_id": name_id,
                    "given_name": name,
                    "given_name_index": idx,
                }
                table_records.append(record)
        else:
            record = {
                "name_id": name_id,
                "given_name": None,
                "given_name_index": 0,
            }
            table_records.append(record)
        return table_records

    def _insert_external_person_id(
        self,
        person_id: str,
        external_person_id: str,
    ):
        """
        Inserts a new external person id record into the MPI if the external_person_id
        does not already exist in the MPI.

        :param person_id: The person_id matching the patient record if a match has been
            found in the MPI.
        :param external_person_id: The external_person_id for the patient record if it
            exists.

        """
        if person_id is None or external_person_id is None:  # pragma: no cover
            raise ValueError("person_id and external_person_id must be provided.")

        external_source_id = self._get_external_source_id("IRIS")

        if external_source_id is not None:
            query = select(self.dal.EXTERNAL_PERSON_TABLE).where(
                text(
                    f"{self.dal.EXTERNAL_PERSON_TABLE.name}.external_person_id"
                    + f" = '{external_person_id}' AND "
                    + f"{self.dal.EXTERNAL_PERSON_TABLE.name}.person_id = '{person_id}'"
                    + f" AND {self.dal.EXTERNAL_PERSON_TABLE.name}.external_source_id "
                    + f" = '{external_source_id}'"
                )
            )
            external_person_record = self.dal.select_results(query, False)

            if len(external_person_record) == 0:
                new_external_person_record = {
                    "person_id": person_id,
                    "external_person_id": external_person_id,
                    "external_source_id": external_source_id,
                }
                self.dal.bulk_insert_list(
                    self.dal.EXTERNAL_PERSON_TABLE, [new_external_person_record], False
                )

    @cache
    def _get_external_source_id(self, external_source_name: str) -> Literal[str, None]:
        """
        Gets the external source id for the external source name provided.
        :param external_source_name: The external source name.
        :return: The external source id if found, otherwise None.
        """

        external_source_id_query = select(self.dal.EXTERNAL_SOURCE_TABLE).where(
            text(
                f"{self.dal.EXTERNAL_SOURCE_TABLE.name}.external_source_name"
                + f" = '{external_source_name}'"
            )
        )
        external_source_record = self.dal.select_results(
            external_source_id_query, False
        )

        external_source_id = None
        if len(external_source_record) > 0:
            external_source_id = external_source_record[0][0]

        return external_source_id

    def _generate_dict_record_from_results(
        self, results_list: List[list]
    ) -> List[dict]:
        """
        Converts a list of list of records into a
        dictionary using the column name header, in the
        first row (first list) along with the values in
        the rest of the rows (lists).


        :param results_list: a list of list containing mpi
            records.
        :return: A dictionary that has the key value pairs
            of the results list.
        """
        return_records = []
        # we must ensure that there is a header AND at least
        # one record or there is much of a point in moving forward
        if len(results_list) > 1 and len(results_list[0]) > 0:
            for row_index, record in enumerate(results_list):
                if row_index > 0:
                    columns_and_values = {}
                    for col_index, row_header in enumerate(results_list[0]):
                        columns_and_values[row_header] = record[col_index]
                    return_records.append(columns_and_values)
        return return_records

    def _insert_person(self) -> str:
        """
        Simple insert of a new person record, which contains
        a new id (pk)

        :return: The newly created person id.
        """
        person_record = {}
        person_id = self.dal.bulk_insert_list(
            self.dal.PERSON_TABLE, [person_record], True
        )
        return person_id[0]



from datetime import date
from datetime import datetime
from typing import Union

from phdi.linkage.config import get_settings


def load_mpi_env_vars_os():
    """
    Simple helper function to load some of the environment variables
    needed to make a database connection as part of the DB migrations.
    """
    dbsettings = {
        "dbname": get_settings().get("mpi_dbname"),
        "user": get_settings().get("mpi_user"),
        "password": get_settings().get("mpi_password"),
        "host": get_settings().get("mpi_host"),
        "port": get_settings().get("mpi_port"),
        "db_type": get_settings().get("mpi_db_type"),
    }
    return dbsettings


# TODO:  Not sure if we will need this or not
# leaving in utils for now until it's determined that
# we won't need to use this within any of the DAL/MPI/LINK
# code
# # https://kb.objectrocket.com/postgresql
# /python-error-handling-with-the-psycopg2-postgresql-adapter-645
# def print_psycopg2_exception(err):
#     # get details about the exception
#     err_type, _, traceback = sys.exc_info()

#     # get the line number when exception occured
#     line_num = traceback.tb_lineno

#     # print the connect() error
#     print("\npsycopg2 ERROR:", err, "on line number:", line_num)
#     print("psycopg2 traceback:", traceback, "-- type:", err_type)

#     # psycopg2 extensions.Diagnostics object attribute
#     print("\nextensions.Diagnostics:", err.diag)

#     # print the pgcode and pgerror exceptions
#     print("pgerror:", err.pgerror)
#     print("pgcode:", err.pgcode, "\n")


def datetime_to_str(
    input_date: Union[str, date, datetime], include_time: bool = False
) -> str:
    """
    Convert a date or datetime object to a string; if a string is provided,
    check that it follows the appropriate format. If unable to perform actions,
    return input as string rather than failing loudly.

    :param input_date: The input date to convert, which prefers types of
        datetime.date, datetime.datetime, or str.
    :param include_time: Whether to include the time in the output string.
    :return: The formatted date as a string. If include_time is True, the
        format is 'YYYY-MM-DD HH:MM:SS', otherwise it's 'YYYY-MM-DD'. If
        empty or None, return empty or None.
    """
    # Handle None or empty string
    if input_date is None or input_date == "":
        return input_date

    # if input is str try to check that it follows the expected format
    if isinstance(input_date, str):
        try:
            expected_format = "%Y-%m-%d %H:%M:%S" if include_time else "%Y-%m-%d"
            datetime.strptime(input_date, expected_format)
            return input_date
        except ValueError:
            # rather than break loudly, allow str to pass
            return input_date

    # if input is a date or datetime then convert in the expected format
    elif isinstance(input_date, (date, datetime)):
        if include_time:
            return input_date.strftime("%Y-%m-%d %H:%M:%S")
        else:
            return input_date.strftime("%Y-%m-%d")
    # if input isn't any of the accepted formats, then return a type error
    else:
        try:
            return str(input_date)
        except TypeError:
            raise TypeError(
                f"Input date {input_date} is not of type date, datetime, "
                "or str; or, it can't be converted or returned safely."
            )
